import math
import os
import random
import pygame
import time

DEBUG_MODE = False
SOUND_ENABLED = True

if not DEBUG_MODE:
	SOUND_ENABLED = True

GO_TO_TITLE = DEBUG_MODE
SHOW_FRAME_DROP = DEBUG_MODE

def getSpecialLevelStuff(id, scene):
	if id == 'main':
		return [BalloonPopping(scene, 'water1'),
			BalloonPopping(scene, 'water4'),
			BalloonPopping(scene, 'lava1')]
	if id == 'ship_1':
		return [ShipTransmission(scene)]
	if id == 'grav_core':
		return [GravityCorePlacement(scene)]
	if id == 'grav_ascent':
		return [GravityAscentExit(scene)]
	if id == 'volcano_1_core':
		return [VolcanoCoreActivation(scene, 1)]
	if id == 'volcano_2_core':
		return [VolcanoCoreActivation(scene, 2)]
	if id == 'volcano_3_core':
		return [VolcanoCoreActivation(scene, 3)]
	if id == 'balloon_2_overland':
		return [BalloonPopping(scene, 'water2')]
	if id == 'balloon_3_overland':
		return [BalloonPopping(scene, 'water3')]
	if id == 'bike_level':
		return [BikeLevelStuff(scene)]
	if id == 'volcano_2_overland':
		return [BalloonPopping(scene, 'lava2')]
	if id == 'volcano_3_overland':
		return [BalloonPopping(scene, 'lava3')]
	if id == 'birdfeeder':
		return [BirdFeeder(scene)]
	return []

class SpecialLevelStuff:
	def __init__(self, scene):
		self.scene = scene
		self.context = scene.context
		self.tiles = scene.tiles
		self.hasUpdate = False
		self.hasPostInit = False
		self.hasDoorTrigger = False
		self.hasRender = False
		
		# meh, hacks
		self.lavaLevel = 0
		self.shakeScreen = False
		self.freeze = False
	
	# override and return None if you want that door to do something special. 
	# The doorId will be post-swapped.
	def doorTrigger(self, doorId):
		return doorId

# Add maps to this list to alter what background they'll have

CAVE_BG = [
	'grav_ascent',
	'grav_core',
	'grav_descent',
	'grav_descent_secret',
	'grav_gate',
	'section_3_gate',
	'section_3_gate_active',
  'exit_water_phase_locked',
  'exit_water_phase_unlocked',
  'water_gate_1_locked',
  'water_gate_1_unlocked',
  'wlink_A1',
  'wlink_A2',
  'wlink_A3',
  'wlink_A4',
  'wlink_B_nowater',
  'wlink_B_somewater',
  'wlink_B_allwater',
  'wlink_D1_water',
  'wlink_D1_nowater',
  'wlink_D2_water',
  'wlink_D2_nowater',
  'wlink_D3_water',
  'wlink_D3_nowater',
  'wlink_E1_water',
  'wlink_E1_nowater',
  'wlink_E2',
  'wlink_E3',
  'bird_link',
]

VOLCANO_BG = [
	'volcano_1_ascent',
	'volcano_1_core',
	'volcano_1_core_active',
	'volcano_1_descent',
	'lava_gate_2',
	'lava_gate_2_active',
	'lava_gate_3',
	'lava_gate_3_active',
	'volcano_2_core',
	'volcano_2_lower_ascent',
	'volcano_2_lower_descent',
	'volcano_2_upper_ascent',
	'volcano_2_upper_descent',
	'volcano_2_upper_secret',
	'volcano_3_core',
	'volcano_3_top',
	'volcano_3_middle',
	'volcano_3_bottom',
	'volcano_3_secret'
]

STARS_BG = [
	'rope_test',
	'ship_1'
]





##########################################

_bg = {}
for cbg in CAVE_BG:
	_bg[cbg] = 'cave'
for vbg in VOLCANO_BG:
	_bg[vbg] = 'volcano'
for sbg in STARS_BG:
	_bg[sbg] = 'stars'


def getBackground(level):
	level = level.split('.')[0]
	return _bg.get(level, 'sky')


FULL_LIFE = 10
START_LIFE = 10
START_LIVES = 3

ALPHABET = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

class Context:
	
	def __init__(self):
		self.lifemeter = START_LIFE
		self.lives = START_LIVES
		self.volcanoA = False
		self.volcanoB = False
		self.volcanoC = False
		self.balloonA = False
		self.balloonB = False
		self.balloonC = False
		self.balloonD = False
		self.gravity = False
		self.transmission1 = False
		self.transmission2 = False
		self.transmission3 = False
		self.transmission4 = False
		self.powerupsTaken = {}
	
	def convertToPassword(self):
		
		if self.balloonA or self.balloonB or self.balloonC or self.balloonD:
			count = 3
		elif self.volcanoA or self.volcanoB or self.volcanoC:
			count = 2
		elif self.gravity:
			count = 1
		else:
			count = 0
		
		password = ''
		
		if count == 1:
			password = 'A'
		elif count == 2:
			password = 'B'
		elif count == 3:
			password = 'C'
		else:
			password = self.randomChar(ALPHABET[3:])
		
		if count >= 1:
			if self.gravity:
				password += self.randomChar(ALPHABET[:13])
			else:
				password += self.randomChar(ALPHABET[13:])
		else:
			password += self.randomChar(ALPHABET)
		
		if count >= 2:
			value = self.volcanoA + (self.volcanoB * 2) + (self.volcanoC * 4)
			letter = chr(ord('1') + value - 1)
			password += letter
		else:
			password += self.randomChar('0123456789')
		
		if count >= 3:
			value = self.balloonA + (self.balloonB * 2) + (self.balloonC * 4) + (self.balloonD * 8)
			letter = chr(ord('A') + value - 1)
			password += letter
		else:
			password += self.randomChar(ALPHABET)
		
		return password
		
		
	# First letter:
	# A - 2nd letter is password, 3rd and 4th are random
	# B - 2nd and 3rd letter are password, 4th is random
	# C - All 3 are used
	# D+ - password is random, no state
	
	# Second letter when not random
	# <= M - gravity is on
	# >= N - gravity is off
	
	# Third letter when not random
	# 1 - volcanos 1, 0, 0
	# 2 - volcanos 0, 1, 0
	# 3 - volcanos 1, 1, 0
	# 4 - volcanos 0, 0, 1
	# 5 - volcanos 1, 0, 1
	# 6 - volcanos 0, 1, 1
	# 7 - volcanos 1, 1, 1
	# >= 8 - volcanos 0, 0, 0
	
	# Fourth letter when not random
	# same pattern for volcanos, but for water, also uses letters instead of numbers
	
	def randomChar(self, characters):
		return characters[int(random.random() * len(characters))]
	
	def enforceLetter(self, letter):
		letter = letter.upper()
		if ord(letter) < ord('A'):
			return 'A'
		if ord(letter) > ord('Z'):
			return 'Z'
		return letter
	
	def enforceNumber(self, char):
		if ord(char) < ord('0'):
			return '0'
		if ord(char) > ord('9'):
			return '9'
		return char
	
	def convertFromPassword(self, password):
		p1 = self.enforceLetter(password[0])
		p2 = self.enforceLetter(password[1])
		p3 = self.enforceNumber(password[2])
		p4 = self.enforceLetter(password[3])
		
		if p1 in 'ABC':
			if ord(p2) <= ord('M'):
				self.gravity = True
		
		if p1 in 'BC':
			value = ord(p3) - ord('1') + 1
			print(value, ord(p3))
			self.volcanoA = (value & 1) != 0
			self.volcanoB = (value & 2) != 0
			self.volcanoC = (value & 4) != 0
		
		if p1 == 'C':
			value = ord(p4) - ord('A') + 1
			self.balloonA = (value & 1) != 0
			self.balloonB = (value & 2) != 0
			self.balloonC = (value & 4) != 0
			self.balloonD = (value & 8) != 0
		
		self.lifemeter = START_LIFE
		self.lives = START_LIVES
	
	def getCurrentCountAndType(self):
		vcount = self.volcanoA + self.volcanoB + self.volcanoC
		wcount = self.balloonA + self.balloonB + self.balloonC + self.balloonD
		if self.gravity:
			if vcount == 3:
				if wcount == 4:
					return None # bird seed
				return ('W', wcount)
			return ('V', vcount)
		return None # gravity
	
	def adjustHealth(self, scene, amount):
		self.lifemeter += amount
		if self.lifemeter <= 0:
			self.lifemeter = 0
			scene.next = DeathOverrideScene(scene, 'collapse')
		else:
			if amount < 0:
				if self.lifemeter < 4:
					playNoise('low_health')
				else:
					playNoise('get_hit')
			else:
				playNoise('raise_health')
		
		if self.lifemeter > 10:
			self.lifemeter = 10



cred_BO = "BLAKE O'HARE"
cred_TB = "TED BURTON"
cred_SC = "STEVE CRAWFORD"
cred_LF = "LAURA FREER"
cred_CS = "CHRISTINE SANDQUIST"
cred_AM = "ANGEL MCLAUGHLIN"
cred_AC = "ADRIAN CLINE"



CREDITS = [
	('TEAM NERD PARADISE', []),
	('PROGRAMMING', [cred_BO]),
	('LEVEL DESIGN', [cred_TB, cred_SC, cred_LF]),
	('CHARACTER ART', [cred_CS]),
	('LEVEL ART', [cred_AM]),
	('MUSIC AND SFX', [cred_AC]),
	('DIALOG', [cred_LF]),
	('AND YOU!', ['FOR PLAYING', 'OR SOMETHING LIKE THAT']),
	('', ['THE END'])
]

class CreditsScene:
	def __init__(self):
		self.next = self
		self.flags = 'M'
		self.counter = 0
		
		images = []
		y = 0
		
		i = 0
		while i < len(CREDITS):
			images.append((getText(WHITE, CREDITS[i][0]), y))
			
			y += 16
			for person in CREDITS[i][1]:
				images.append((getText(WHITE, person), y))
				y += 8
			
			y += 64
			
			i += 1
		self.images = images
	
	def processInput(self, events, pressed):
		if self.counter > 900:
			for event in events:
				if event.down:
					if event.action in ('start', 'A', 'B'):
						self.next = TitleScene()
	
	def update(self):
		self.playMusic()
		self.counter += 1
	
	def render(self, screen, rc):
		screen.fill(BLACK)
		
		counter = min(self.counter, 892)
		
		for image in self.images:
			screen.blit(image[0], (64, image[1] - counter + 224))
	
	def playMusic(self):
		JUKEBOX.ensureSong('credits')

class DeathOverrideScene:
	# type = "lava", "collapse", "fall"
	def __init__(self, bg, type):
		self.bg = bg
		self.context = bg.context
		self.context.lifemeter = 0
		self.flags = ''
		self.next = self
		self.type = type
		self.counter = 0
		self.bg.player.deathState = type
		self.vy = -10
		if type == 'fall':
			self.vy = bg.player.vy
		elif type == 'lava':
			playNoise('lava_roast')
		
		
	def processInput(self, events, pressed):
		pass # There is nothing you can do
	
	def update(self):
		
		if self.counter == 0:
			JUKEBOX.ensureSong('death')
			
		self.counter += 1
		self.vy += 1
		if self.type in ('lava', 'fall'):
			self.bg.player.modelY += self.vy
			self.bg.player.y = int(self.bg.player.modelY)
			
		if self.counter > 100:
			self.context.lifemeter = 10
			self.context.lives -= 1
			if self.context.lives == 0:
				self.context.lives = 3
				self.next = GameOverScene(self.context)
			else:
				self.next = ReadyScene(self.bg, self.context)
			# TODO: game over or beginning of same map at same door entrance
		
	
	def render(self, screen, rc):
		self.bg.render(screen, rc)
	

class DeathScene:
	def __init__(self):
		print("Death!")
		self.next = None
		self.flags = ''
	
	def processInput(self, events, pressed):
		pass
	
	def update(self):
		pass
	
	def render(self, screen, rc):
		pass
		

DIALOGS = {

	'Feeder1': [
		["Quick, collect seeds",
		 "before anyone notices!"]
	],
	
	"Feeder2": [
		["Uh-oh, your covert",
		 "attempt has failed!",
		 "Quick, make a getaway!"]
	],

	'WaterDone': [
		["That's the last of the",
		 "reservoirs. Receiving",
		 "Transmission from the",
		 "General..."],
		
		["We have a small hiccup;",
		"the seeds intended to",
		"provide vegetation were",
		"accidentally left here."],

		["Fortunately, a vanguard",
		"of our sworn enemies is",
		"north of your location",
		"bearing their own seeds."],

		["You can simultaneously",
		"stop their invasion and",
		"steal their seeds for",
		"our own purposes."],

		["It's a case of killing",
		"two birds with one",
		"stone! Or anything else",
		"you might prefer."]
	],
	
	'GravityDone': [
		["The gravity emitter is",
		 "now taking effect."],
		
		["You can now hold B to",
		 "run and jump."],
		
		["Running jumps reach",
		 "higher spots."],
		
		["Head back to the ship",
		 "for further instructions",
		 "from the General."]
	],

	'Open1': [
		["An intrepid explorer",
		 "left home in search of",
		 "a better life for all",
		 "squirrelkind."],
		
		["Bearing the hopes and",
		 "dreams of her people,",
		 "the squirrel set course",
		 "for a nearby moon."],
	],
	
	'Open2': [
		["Her vessel reaches the",
		 "moon laden with large",
		 "supply stores as well as",
		 "terraforming equipment"],
		
		["...including the very",
		 "latest in hydration",
		 "technology:"],
		
		["Giant water balloons."],
	],
	
	'Open3': [
		["As she gazes upon the",
		 "barren waste of the",
		 "desolate moon, she",
		 "says to herself:"],
		
		["`I've got a lot of",
		 "work to do.\""]
	],

	'Volcano': [
		["That's the last of the",
		 "volcanoes."],
		 
		["Now that the magnetic",
		 "field is back, the",
		 "atmosphere will",
		 "re-accumulate."]
	],
	
	'T1': [
		["Hello explorer! We trust",
		"your journey was a",
		"pleasant one."],

		["Your first task is to",
		"normalize the gravity on",
		"the moon."],
	
		["Place the gravity",
		"emitter within the",
		"moon's core."],
	
		["Remember we're relying",
		"on you; don't forget the",
		'`gravity" of our',
		"situation!"]
	],
	
	
	'T2': [
		["Now in order to restore",
		"the moon's magnetic",
		"field, you'll need some",
		"specialized equipment."],

		["Take your packet of",
		"Acme Instant Lava and",
		"sprinkle a little into",
		"inactive volcanoes."],
		
		["Presto, lava! And as",
		"everyone knows, lava",
		"creates magnetic fields."],

		["SCIENCE!"]
	],

	'T3': [
		["When you landed, your",
		"H2O reservoirs should",
		"have been released."],
		
		["You will need to break",
		"the rubber coating on",
		"the reservoirs to",
		"release the water."]
	],
	
	'T4': [
		["We have a small hiccup;",
		"the seeds intended to",
		"provide vegetation were",
		"accidentally left here."],

		["Fortunately, a vanguard",
		"of our sworn enemies is",
		"heading to your location",
		"bearing their own seeds."],

		["You can simultaneously",
		"stop their invasion and",
		"steal their seeds for",
		"our own purposes."],

		["It's a case of killing",
		"two birds with one",
		"stone! Or anything else",
		"you might prefer."]
	],
	
	'CloseDialog': [
		["Congratulations!",
		 "The Space Squirrel has",
		 "managed to bring life",
		 "to this once dead moon."],
		 
		["Now her people have a",
		 "beautiful new home, and",
		 "it's all thanks to you!"]
	]
}

tooBig = False
for key in DIALOGS.keys():
	for stanza in DIALOGS[key]:
		for line in stanza:
			if len(line) > 24:
				print("The following line is too long in conversation '" + key + "':")
				print('    >',line)
				tooBig = True

if tooBig:
	print("Will divide by zero and crash until Satyrane fixes it.")
	z = 1 / 0 # error
				



class DialogScene:
	def __init__(self, bg, dialogId, slightDelay, storyMode=False, forceArrow=False):
		self.next = self
		self.flags = ''
		self.bg = bg
		self.storyMode = storyMode
		self.forceArrow = forceArrow
		player = bg.player
		if player != None:
			player.moving = False
			player.lastDirection = 'left'
		self.delayCounter = 45 if slightDelay else 0
		self.stanzas = DIALOGS[dialogId]
		self.stanzaIndex = 0
		self.textCursor = 0
		self.showLines = []
		self.blink = False
		self.id = dialogId
		
		
		 # The epitome of a PyWeek hack
		self.generalLoc = None
		if len(dialogId) == 2 and dialogId[0] == 'T':
			self.generalLoc = (4 * 16, 6 * 16)
	
	def processInput(self, events, pressed):
		if self.blink:
			for event in events:
				if event.down and event.action in ('A', 'B', 'start'):
					self.stanzaIndex += 1
					self.textCursor = 0
		else:
			if pressed['A'] or pressed['B'] or pressed['start']:
				if self.delayCounter < 0:
					self.textCursor += 1.2
		
	
	def update(self):
	
		if 'M' in self.bg.flags:
			self.bg.playMusic()
		
		self.delayCounter -= 1
		if self.stanzaIndex >= len(self.stanzas):
			self.next = self.bg
			self.bg.next = self.bg
			
			# Total hack...
			#if self.id == 'WaterDone':
			#	self.next = PlayScene('ship_1', 8, 9, self.bg.context)
			
		else:
			self.activeStanza = self.stanzas[self.stanzaIndex]
			rawLines = '\n'.join(self.activeStanza)
			self.blink = False
			cursor = int(self.textCursor)
			if cursor >= len(rawLines):
				self.blink = True 
			else:
				rawLines = rawLines[:cursor]
			showLinesStr = rawLines.split('\n')
			showLines = []
			for line in showLinesStr:
				if len(line) > 0:
					showLines.append(getText((255, 255, 255), line))
			self.showLines = showLines
		
		if self.delayCounter < 0:
			self.textCursor += 0.3
	
	def render(self, screen, rc):
		self.bg.render(screen, rc)
		
		if self.delayCounter > 0: return
		
		if self.generalLoc != None:
			path = 'tiles/screen-static/screen' + ('', '1', '2', '3')[(rc // 4) & 3] + '.png'
			screen.blit(getImage(path), self.generalLoc)
		
		if not self.storyMode:
			drawBox(screen, 16, 16, 28, 9)
		x = 32
		y = 32
		space = 8
		if self.storyMode:
			y += 16 * 7
			space = 16
		
		startY = y
		for line in self.showLines:
			screen.blit(line, (x, y))
			y += space
		if self.blink and ((rc // 7) & 1) == 0:
			pos = (256 - 40, startY + space * 4)
			if not self.forceArrow and self.stanzaIndex == len(self.stanzas) - 1:
				pygame.draw.rect(screen, (255, 255, 255), pygame.Rect(pos[0], pos[1], 8, 8))
			else:
				screen.blit(getText((255, 255, 255), '^'), pos)
		
		if not self.blink and (rc % 5 == 0):
			playNoise('text_char')

ACORN_SPEED = 1.5
ACORN_DELAY_FOR_TOP_THROWING = 5 * 30
class AcornAutomation:
	def __init__(self, sprite):
		self.sprite = sprite
		self.countdown = int(random.random() * ACORN_DELAY_FOR_TOP_THROWING)
		self.mode = 'patrolling' # patrolling | throwing | baldpatrolling
		self.top = None
		self.walkingLeft = True
	
	def doStuff(self, scene):
		self.countdown -= 1
		walk = False
		if self.mode == 'patrolling':
			if self.countdown <= 0:
				self.mode = 'throwing'
				self.top = Sprite('acorntop', self.sprite.x, self.sprite.y)
				self.top.acorntopdir = self.sprite.lastDirection
				self.top.goLeft = self.top.acorntopdir == 'left'
				self.top.automation.setGoLeft(self.sprite.lastDirection == 'left')
				self.top.automation.setGoUp(self.sprite.y > scene.player.y)
				self.top.automation.body = self.sprite
				self.top.deleteWhenOffScreen = True
				self.top.ghost = True
				self.top.floats = True
				self.sprite.spawns = [self.top]
			else:
				walk = True
		elif self.mode == 'throwing':
			if self.top.dead:
				self.mode = 'baldpatrolling'
			elif self.top.automation.completed:
				self.mode = 'patrolling'
				self.top.dead = True
				self.countdown = ACORN_DELAY_FOR_TOP_THROWING
		elif self.mode == 'baldpatrolling':
			walk = True
		
		if walk:
			if self.sprite.collidedWall:
				self.walkingLeft = not self.walkingLeft
			
			tiles = scene.tiles
			cols = scene.cols
			rows = scene.rows
			dx = ACORN_SPEED if self.walkingLeft else -ACORN_SPEED
			oldTileX = int(self.sprite.modelX / 16)
			newX = self.sprite.modelX + dx
			oldTileY = int(self.sprite.modelY / 16)
			newTileX = int(newX / 16)
			if newTileX < 0 or newTileX >= cols:
				self.walkingLeft = not self.walkingLeft
				return
			
			belowTileY = oldTileY + 1
			
			if newTileX == oldTileX:
				self.sprite.dx = dx
			else:
				if belowTileY < 0 or belowTileY >= rows:
					pass
				t = tiles[newTileX][belowTileY]
				if not t.solid and not t.isTop:
					self.walkingLeft = not self.walkingLeft
					pass
				else:
					self.sprite.dx = dx
			
			# TODO: walk back and forth, but don't fall off.


ACORN_TOP_SPEED = 4.5

TRAIL = []

# this is the best I can come up with off the top of my head. I'm sure there's a faster way to create a parametric function of the acorn top path

ball = [0, 0]

waypoints = [
	[-1, 0],
	[-6, 0],
	[-6.6, -0.4],
	[-7, -1],
	[-6.6, -1.6],
	[-6, -2],
	[-1, -2],
	[-0.4, -1.6],
	[0, -1],
	[0, -.5]]

v = ACORN_TOP_SPEED + 0.0

i = 0
while i < len(waypoints):
	
	target = waypoints[i]
	tx = target[0] * 16.0
	ty = target[1] * 16.0
	
	done = False
	while not done:
		dx = tx - ball[0]
		dy = ty - ball[1]
		distance = (dx ** 2 + dy ** 2) ** .5
		if distance <= v:
			done = True
			TRAIL.append((tx, ty))
			ball = [tx, ty]
		else:
			dx *= v / distance
			dy *= v / distance
			ball[0] += dx
			ball[1] += dy
			TRAIL.append((int(10 * ball[0]) / 10.0, int(10 * ball[1]) / 10.0))
	i += 1


class AcornTopAutomation:
	def __init__(self, sprite):
		self.sprite = sprite
		self.trail = TRAIL[:]
		self.trailIndex = 0
		self.startX = sprite.modelX
		self.startY = sprite.modelY
		self.completed = False
	
	def setGoLeft(self, goLeft):
		self.goLeft = goLeft
		self.xSign = 1 if self.goLeft else -1
	
	def setGoUp(self, goUp):
		self.ySign = 1 if goUp else -1
	
	def doStuff(self, scene):
		if self.trailIndex < len(self.trail):
			
			rawX = self.trail[self.trailIndex][0]
			rawY = self.trail[self.trailIndex][1]
			
			rawX *= self.xSign
			rawY *= self.ySign
			
			targetX = self.startX + rawX
			targetY = self.startY + rawY
			
			dx = targetX - self.sprite.modelX
			dy = targetY - self.sprite.modelY
			
			self.sprite.dx = dx
			self.sprite.dy = dy
			
			self.trailIndex += 1
		else:
			if self.completed:
				self.sprite.dead = True
			self.completed = True
				

class BirdAutomation:
	def __init__(self, sprite, onBikeLevel):
		sprite.isEnemy = True
		self.counter = 0
		self.mode = 'stalk' # 'right' | 'left'
		self.sprite = sprite
		self.bikeLevel = onBikeLevel
		
	def doStuff(self, scene):
		self.counter += 1
		p = scene.player
		sprite = self.sprite
		if self.mode == 'stalk':
			dy = p.y - self.sprite.y
			
			if abs(dy) < 6:
				self.mode = 'right'
			elif dy > 0:
				self.sprite.dy += 3
			else:
				self.sprite.dy -= 3
		elif self.mode == 'right':
			if sprite.x < p.x:
				sprite.dx = 6
			else:
				self.mode = 'left'
		else: # mode == 'left'
			spriteX = sprite.x
			if self.bikeLevel:
				spriteX -= self.counter * BIKE_SPEED
			
			if spriteX > 7:
				sprite.dx = -6
			else:
				self.mode = 'stalk'
		
		if self.bikeLevel:
			sprite.dx += BIKE_SPEED

class ByatAutomation:
	def __init__(self, sprite):
		self.sprite = sprite
		self.perched = False
		self.counter = 0
		self.mode = 'perch'
		self.perched = True
		self.dropHeight = 10
		
	def doStuff(self, scene):
		self.counter += 1
		sprite = self.sprite
		tiles = scene.tiles
		
		if self.mode == 'perch':
			self.perched = True
			if self.counter > 3 * 30:
				self.counter = 0
				self.mode = 'down'
				self.perched = False
				dropHeight = int(random.random() * 5 + 1) * 16
				self.targetY = sprite.y + dropHeight
				#self.originalY = sprite.y
				
		elif self.mode == 'down':
			sprite.dy = 2
			if sprite.y >= self.targetY or sprite.collidedWall or sprite.onGround:
				self.mode = 'swoop'
				self.counter = 0
				self.swoopLeft = scene.player.x < sprite.x
		elif self.mode == 'swoop':
			swoopSpeed = 1.6
			sprite.dy = math.sin(self.counter * 40 / (2 * 3.14159)) * 4
			sprite.dx = -swoopSpeed if self.swoopLeft else swoopSpeed
			if self.counter > 30 * 6:
				self.mode = 'up'
				self.counter = 0
			elif sprite.collidedWall:
				self.swoopLeft = not self.swoopLeft
		
		elif self.mode == 'up':
			sprite.dy = -3.5
			mt = sprite.x & 15
			if mt < 4:
				sprite.dx = 1
			elif mt > 12:
				sprite.dx = -1
			stop = False
			if self.sprite.collidedWall:
				stop = True
			else:
				tx = sprite.x // 16
				ty = (sprite.y - 8) // 16
				if tx < 0 or tx >= scene.cols or ty < 0 or ty >= scene.rows:
					stop = True
				else:
					t = tiles[tx][ty]
					if t.isTop or t.solid:
						stop = True
			
			if stop:
				self.mode = 'perch'
				self.counter = 0
		

FIREBALL_VELOCITY = 1.5


class FireBallAutomation:
	def __init__(self, sprite):
		self.sprite = sprite
		self.counter = 0
		self.startX = sprite.x
		self.startY = sprite.y
		self.goLeft = sprite.lastDirection == 'left'
		self.vx = -FIREBALL_VELOCITY
		self.vy = 0
	
	def doStuff(self, scene):
		if self.counter == 0:
			dx = scene.player.x - self.startX
			dy = scene.player.y - self.startY
			
			if dx == 0:
				dx = 1
							
			m = 1.0 * dy / dx
			if m > 1.0:
				m = 1.0
			elif m < -1.0:
				m = -1.0
			ang = math.atan(m)
			self.vx = FIREBALL_VELOCITY * math.cos(ang)
			self.vy = -FIREBALL_VELOCITY * math.sin(ang)
			if self.goLeft:
				self.vx *= -1
		else:
			if self.predictedX != self.sprite.x:
				self.sprite.dead
				
		self.sprite.dx += self.vx
		self.sprite.dy += self.vy
		
		self.predictedX = int(self.sprite.dx + self.sprite.modelX)
		
		if self.sprite.collidedWall:
			self.sprite.dead = True
		
				
			
			
		self.counter += 1

RISE_SPEED = 1
WAIT_INTERVAL = 30 * 4

class LavaMonsterAutomation:
	def __init__(self, sprite):
		self.sprite = sprite
		self.counter = 0
		self.mode = 'wait' # { wait | rise | shoot | fall }
		self.mouthOpen = False
		
	
	def doStuff(self, scene):
		self.counter += 1
		self.mouthOpen = False
		faceLeft = scene.player.x < self.sprite.x
		self.sprite.lastDirection = 'left' if faceLeft else 'right'
		if self.mode == 'wait':
			if self.counter == WAIT_INTERVAL:
				self.counter = 0
				self.mode = 'rise'
		elif self.mode == 'rise':
			if self.counter == 16:
				self.mode = 'shoot'
				self.counter = 0
			self.sprite.dy = -1
		elif self.mode == 'shoot':
			if self.counter == 60:
				self.mode = 'fall'
				self.counter = 0
			if self.counter == 7:
				self.createFireBall()
			if self.counter > 0 and self.counter < 14:
				self.mouthOpen = True
		
		elif self.mode == 'fall':
			if self.counter == 16:
				self.mode = 'wait'
				self.counter = 0
			self.sprite.dy = 1
		
		
	def createFireBall(self):
		fireball = Sprite('fireball', self.sprite.x, self.sprite.y)
		self.sprite.spawns = [fireball]
		


class LazorAutomation:
	def __init__(self, sprite):
		self.sprite = sprite
		
	def doStuff(self, scene):
		vx = self.sprite.bvx
		self.sprite.dx = self.sprite.bvx
		if scene.id == 'bike_level':
			self.sprite.dx += 5

CHASE_DISTANCE_TILES = 6

class MoonSquidAutomation:
	def __init__(self, sprite):
		self.sprite = sprite
		self.counter = 0
		self.goLeft = random.random() < .5
	
	def doStuff(self, scene):
		self.counter += 1
		player = scene.player
		sprite = self.sprite
		dx = player.x - sprite.x
		dy = player.y - sprite.y
		if dx * dx + dy * dy < (16 * CHASE_DISTANCE_TILES) ** 2:
			v = (math.sin(self.counter / 30.0 * 2 * 3.14159 / 2) + 1) / 1.5
			if dx == 0 and dy == 0:
				pass
			else:
				d = (dx * dx + dy * dy) ** .5
				dx *= v / d
				dy *= v / d
				sprite.dx = dx
				sprite.dy = dy
		else:
			if sprite.collidedWall:
				self.goLeft = not self.goLeft
			else:
				sprite.dx =  -1 if self.goLeft else 1


class PoofAutomation:
	def __init__(self, sprite, isBig):
		self.sprite = sprite
		self.counter = 0
		self.isBig = isBig
		
	
	def doStuff(self, scene):
		self.counter += 1
		if self.counter == 15:
			self.sprite.dead = True

class PowerupAutomation:
	def __init__(self, sprite):
		self.sprite = sprite
		self.counter = 0
		self.visible = True
		
	def doStuff(self, scene):
		player = scene.player
		if self.sprite.powerupInfo.id == None:
			self.counter += 1
			if self.counter > 8 * 30:
				self.sprite.dead = True
			elif self.counter > 6 * 30:
				self.visible = (self.counter & 1) == 0
			
		for playerY in (player.y, player.y - 16):
			x = player.x
			y = playerY
			dx = self.sprite.x - x
			dy = self.sprite.y - y
			if dx * dx + dy * dy < 144:
				self.applyPowerup(scene.context)
	
	def applyPowerup(self, context):
		self.sprite.dead = True
		info = self.sprite.powerupInfo
		id = info.id
		type = info.type
		if id != None:
			context.powerupsTaken[id] = True
		if type == 'life_small':
			playNoise('raise_health')
			context.lifemeter += 2
			if context.lifemeter > 10:
				context.lifemeter = 10
		elif type == 'life_big':
			playNoise('raise_health')
			context.lifemeter = 10
		elif type == '1up':
			playNoise('1up')
			context.lives += 1


class RocketBikeAutomation:
	def __init__(self, sprite):
		self.counter = 0
		self.sprite = sprite
		self.used = False
		
	def doStuff(self, scene):
		p = scene.player
		dx = p.x - self.sprite.x
		dy = p.y - self.sprite.y
		if dx * dx + dy * dy < 100:
			if scene.special[0].seedsLeft == 0:
				scene.next = PlayScene('bike_level', 8, 7, scene.context)

class SeedAutomation:
	def __init__(self, sprite):
		self.counter = 0
		self.sprite = sprite
		self.used = False
		
	def doStuff(self, scene):
		p = scene.player
		dx = p.x - self.sprite.x
		dy = p.y - self.sprite.y
		if dx * dx + dy * dy < 100 and not self.used:
			self.sprite.dead = True
			self.used = True
			scene.special[0].seedsLeft -= 1
			playNoise('raise_health')

class WaterPopAutomation:
	def __init__(self, sprite):
		self.counter = 0
		self.sprite = sprite
		
	def doStuff(self, scene):
		self.counter += 1
		if self.counter >= 15:
			self.sprite.dead = True

class GameOverScene:
	def __init__(self, context):
		self.next = self
		self.flags = ''
		self.context = context
		self.counter = 0
		self.index = 0
		self.mode = 'prompt'
		self.password = context.convertToPassword()
		context.powerupsTaken = {}
		
	def processInput(self, events, pressed):
		if self.mode == 'prompt':
			for event in events:
				if event.down:
					if event.action == 'up':
						if self.index == 1:
							self.index = 0
							playNoise('menu_move')
					elif event.action == 'down':
						if self.index == 0:
							self.index = 1
							playNoise('menu_move')
					elif event.action in ('A', 'B', 'start'):
						self.mode = 'commit'
						self.counter = 0
						playNoise('menu_select')
	
	def update(self):
		JUKEBOX.ensureSong('gameover')
		self.counter += 1
		if self.mode == 'commit':
			if self.counter > 30:
				if self.index == 0:
					self.next = PlayScene('ship_1', 8, 9, self.context)
				elif self.index == 1:
					self.next = TitleScene()
				else:
					# what?
					self.next = OpeningScene()
		
	
	def render(self, screen, rc):
		screen.fill((0, 0, 0))
		show_selected = self.mode == 'prompt' or (rc % 2) == 0
		gameOver = getText(WHITE, "GAME OVER")
		optionA = getText(WHITE, "CONTINUE")
		optionB = getText(WHITE, "END")
		password = getText(WHITE, "PASSWORD:")
		passwordValue = getText((128, 128, 128), self.password)
		cursor = getText(WHITE, ">")
		
		left = 80
		top = 96
		
		x = left
		y = top
		screen.blit(gameOver, (left, top))
		
		y += 16
		if self.index == 0:
			screen.blit(cursor, (x, y))
		if show_selected or self.index == 1:
			screen.blit(optionA, (x + 16, y))
		
		y += 16
		if self.index == 1:
			screen.blit(cursor, (x, y))
		if show_selected or self.index == 0:
			screen.blit(optionB, (x + 16, y))
		
		y += 48
		x -= 32
		screen.blit(password, (x, y))
		x += password.get_width() + 8
		screen.blit(passwordValue, (x, y))

_imageLibrary = {}
_reversedLibrary = {}

def getImage(path):
	img = _imageLibrary.get(path)
	if img == None:
		alphaImg = pygame.image.load(('images/' + path).replace('/', os.sep).replace('\\', os.sep))
		magenta = pygame.Surface(alphaImg.get_size()).convert()
		magenta.fill((255, 0, 255))
		magenta.blit(alphaImg, (0, 0))
		magenta.set_colorkey((255, 0, 255))
		img = magenta
		_imageLibrary[path] = img
	return img

def getBackwardsImage(path):
	img = _reversedLibrary.get(path)
	if img == None:
		img = getImage(path)
		img = pygame.transform.flip(img, True, False)
		_reversedLibrary[path] = img
	return img

# TODO: intro slide show

class IntroScene:
	def __init__(self):
		self.flags = ''
		self.next = self
		self.player = None # lol, pyweek hack
		
		self.page = 0
		self.pages = [
			getImage('slides/open1.png'),
			getImage('slides/open2.png'),
			getImage('slides/open3.png')
			]
		
		self.counter = 0
	
	def processInput(self, events, pressedKeys):
		pass 
	
	def update(self):
		if DEBUG_MODE:
			self.next = self.getStartScene()
			return
		
		JUKEBOX.ensureSong('intro')
		
		self.counter += 1
		
		if self.counter == 10:
			self.next = DialogScene(self, 'Open1', False, True, True)
		elif self.counter == 15:
			self.page = 1
		elif self.counter == 20:
			self.next = DialogScene(self, 'Open2', False, True, True)
		elif self.counter == 25:
			self.page = 2
		elif self.counter == 30:
			self.next = DialogScene(self, 'Open3', False, True, False)
		elif self.counter == 35:
			self.next = self.getStartScene()
		
	def render(self, screen, rc):
		screen.fill(BLACK)
		screen.blit(self.pages[self.page], (0, 16))
	
	def getStartScene(self):
		if os.path.exists('fast_start.txt'):
			c = open('fast_start.txt', 'rt')
			t = c.read().split('\n')
			c.close()
			i = 0
			ignore = False
			level = None
			x = 2
			y = 2
			flags = []
			for line in t:
				line = trim(line.split('#')[0])
				if len(line) > 0:
					i += 1
					if i == 1:
						if line == 'IGNORE':
							ignore = True
							break
						level = line
					elif i == 2:
						parts = line.split(',')
						x = int(trim(parts[0]))
						y = int(trim(parts[1]))
					else:
						flags.append(trim(line).lower())
			
			if not ignore:
				c = Context()
				for flag in flags:
					if flag == 'gravity':
						c.gravity = True
					elif flag == 'volcano1':
						c.volcanoA = True
					elif flag == 'volcano2':
						c.volcanoB = True
					elif flag == 'volcano3':
						c.volcanoC = True
					elif flag == 'water1':
						c.balloonA = True
					elif flag == 'water2':
						c.balloonB = True
					elif flag == 'water3':
						c.balloonC = True
					elif flag == 'water4':
						c.balloonD = True
				
				return PlayScene(level, x, y, c)
				
		return PlayScene('ship_1', 8, 9, Context())

"""
config is a dictionary of button types as strings to their configuration
configuration entries per button are in the following formats:

BUTTON:
('B', index)
HAT:
('H', index, 0 or 1 for the ordinate, True for positive sign)
AXIS:
('A', index, True for positive sign)
"""

cached_joysticks = {} # NAME to CONFIGURATION

previous_active_joystick = None
joysticks_present = {} # NAME to INDEX
joystick_instances = [] # INDEX to INSTANCE

JOYSTICK_BUTTON_ORDER = 'A B up down left right start'.split(' ')
#def get_joystick_config(

def configureJoysticksFromFile():
	if os.path.exists('js_cache.txt'):
		c = open('js_cache.txt', 'rt')
		t = trim(c.read()).split('\n')
		c.close()
		configs = {}
		first = True
		active = None
		for line in t:
			if first:
				if len(line) > 0:
					active = line
				first = False
			parts = line[:-1].split('|')
			if len(parts) < 8:
				continue
			cfgs = []
			cfgs += [('A', parts[0])]
			cfgs += [('B', parts[1])]
			cfgs += [('up', parts[2])]
			cfgs += [('down', parts[3])]
			cfgs += [('left', parts[4])]
			cfgs += [('right', parts[5])]
			cfgs += [('start', parts[6])]
			name = '|'.join(parts[7:])
			
			configs[name] = {}
			ignoreMe = False
			for i in range(len(cfgs)):
			
				button = cfgs[i][0]
				cfg = cfgs[i][1].split('^')
				type = cfg[0]
				if type == 'B': # button
					num = int(cfg[1])
					cfg = ('B', num)
				elif type == 'H': # hat
					num = int(cfg[1])
					ordinate = int(cfg[2])
					positive = cfg[3] == '+'
					cfg = ('H', num, ordinate, positive)
				elif type == 'A':
					num = int(cfg[1])
					positive = cfg[2] == '+'
					cfg = ('A', num, positive)
				else:
					ignoreMe = True
					break
					# uhhh....
				
				
				configs[name][button] = cfg
			if ignoreMe:
				configs.pop(name)
			else:
				cached_joysticks[name] = configs[name]
		
		if cached_joysticks.get(active, None) != None:
			global previous_active_joystick
			previous_active_joystick = active
		

def getJoysticksPresent():
	for i in range(pygame.joystick.get_count()):
		js = pygame.joystick.Joystick(i)
		js.init()
		name = js.get_name()
		joysticks_present[name] = i
		joystick_instances.append(js)



BUTTONS = 'A B up down left right start'.split(' ')

def set_active_joystick(name):
	global active_joystick
	active_joystick = name

active_joystick = None # joystick NAME
active_joystick_snapshot = {}

for button in BUTTONS:
	active_joystick_snapshot[button] = False

names_cached = {}
def get_joystick_manifest():
	value = names_cached.get('value')
	if value == None:
		value = []
		for js in joystick_instances:
			value.append(js.get_name())
		names_cached['value'] = value
	return value
		

def poll_active_joystick(events_out, pressedActions):
	if active_joystick == None:
		return
	
	name = active_joystick
	index = joysticks_present.get(name)
	if index == None or len(joystick_instances) <= index:
		return
	ajs = active_joystick_snapshot
	js = joystick_instances[index]
	config = cached_joysticks[name]
	for button in JOYSTICK_BUTTON_ORDER:
		cfg = config[button]
		type = cfg[0]
		if type == 'B':
			pressed = not not js.get_button(cfg[1])
			if ajs[button] != pressed:
				ajs[button] = pressed
				events_out.append(MyEvent(button, pressed))
				pressedActions[button] = pressed
		elif type == 'H':
			state = js.get_hat(cfg[1])[cfg[2]]
			if cfg[3]:
				pressed = state == 1
			else:
				pressed = state == -1
			if ajs[button] != pressed:
				ajs[button] = pressed
				events_out.append(MyEvent(button, pressed))
				pressedActions[button] = pressed
		elif type == 'A':
			state = js.get_axis(cfg[1])
			if cfg[2]:
				pressed = state > 0.5
			else:
				pressed = state < -0.5
			if ajs[button] != pressed:
				ajs[button] = pressed
				events_out.append(MyEvent(button, pressed))
				pressedActions[button] = pressed
		else: pass # Destroy the universe and start over. Or just ignore.

def serialize_joystick_config():
	output = []
	if active_joystick == None:
		output.append('')
	else:
		output.append(active_joystick)
	for key in cached_joysticks.keys():
		line = []
		config = cached_joysticks[key]
		for button in 'A B up down left right start'.split(' '):
			cfg = config[button]
			if cfg[0] == 'B':
				line.append('B^' + str(cfg[1]))
			elif cfg[0] == 'H':
				line.append('H^' + str(cfg[1]) + '^' + str(cfg[2]) + '^' + str('+' if cfg[3] else '-'))
			elif cfg[0] == 'A':
				line.append('A^' + str(cfg[1]) + '^' + str('+' if cfg[2] else '-'))
		line.append(key)
		line.append('')
		
		output.append('|'.join(line))
	output = '\n'.join(output)
	c = open('js_cache.txt', 'wt')
	c.write(output)
	c.close()


class JoystickConfigScreen:
	def __init__(self, js_index, bg):
		self.bg = bg
		self.bg.next = self.bg
		self.js_index = js_index
		self.next = self
		self.flags = ''
		self.buttons = 'left right up down A B start'.split(' ')
		self.phase = 0
		self.configs_taken = {}
		self.config = {}
	
	def processInput(self, events, pressed):
		pass
	
	def update(self):
		if self.phase < len(self.buttons):
			if self.do_poll():
				self.phase += 1
		
		if self.phase == len(self.buttons):
			name = joystick_instances[self.js_index].get_name()
			cached_joysticks[name] = self.config
			set_active_joystick(name)
			serialize_joystick_config()
			self.next = TitleScene()
	
	def do_poll(self):		
		button = self.buttons[self.phase]
		js = joystick_instances[self.js_index]
		
		for i in range(js.get_numbuttons()):
			if js.get_button(i):
				key = 'B' + str(i)
				if not self.configs_taken.get(key, False):
					self.configs_taken[key] = True
					self.config[button] = ('B', i)
					return True
		
		for i in range(js.get_numaxes()):
			value = js.get_axis(i)
			if value > 0.5:
				key = "A" + str(i) + '+'
				if not self.configs_taken.get(key, False):
					self.configs_taken[key] = True
					self.config[button] = ('A', i, True)
					return True
			elif value < -0.5:
				key = "A" + str(i) + '-'
				if not self.configs_taken.get(key, False):
					self.configs_taken[key] = True
					self.config[button] = ('A', i, False)
					return True
		
		for i in range(js.get_numhats()):
			value = js.get_hat(i)
			for j in (0, 1):
				if value[j] == 1:
					key = 'H' + str(i) + "|" + str(j) + "+"
					if not self.configs_taken.get(key, False):
						self.configs_taken[key] = True
						self.config[button] = ('H', i, j, True)
						return True
				elif value[j] == -1:
					key = 'H' + str(i) + "|" + str(j) + '-'
					if not self.configs_taken.get(key, False):
						self.configs_taken[key] = True
						self.config[button] = ('H', i, j, False)
						return True
		return False
	
	def render(self, screen, rc):
		self.bg.render(screen, rc)
		drawBox(screen, 40, 64, 20, 15)
		
		if self.phase < len(self.buttons):
			button = self.buttons[self.phase]
			txt = getText((255, 255, 255), "Press " + button)
			screen.blit(txt, (40 + 16, 64 + 16))
			img_name = button.lower()
			if ((rc // 6) % 2) == 0:
				img_name = 'none'
			screen.blit(getImage('misc/nes_' + img_name + '.png'), (40 + 16, 64 + 32))
			

class JoystickMenuScene:
	def __init__(self):
		self.next = self
		self.flags = ''
		self.options = ['None'] + get_joystick_manifest()
		self.index = 0 if (active_joystick == None) else (1 + joysticks_present[active_joystick])
		self.blink_counter = 0
	
	def processInput(self, events, pressedActions):
		for event in events:
			if event.down and self.blink_counter <= 0:
				if event.action == 'up':
					self.index -= 1
				elif event.action == 'down':
					self.index += 1
				elif event.action in ['A', 'B', 'start']:
					if self.index == 0:
						set_active_joystick(None)
						self.next = TitleScene()
					else:
						self.blink_counter = 30
				if self.index < 0:
					self.index = 0
				if self.index >= len(self.options):
					self.index = len(self.options) - 1
				
	
	def update(self):
		self.blink_counter -= 1
		if self.blink_counter == 1:
			self.next = JoystickConfigScreen(self.index - 1, self)
			self.blink_counter = 0
	
	def render(self, screen, rc):
		title = getText((255, 255, 0), "Joystick Selection")
		cursor = getText((255, 255, 255), ">")
		screen.fill((0, 0, 0))
		
		left = 32
		drawBox(screen, left - 16, 16, 30, (len(self.options) + 1) * 2 + 3)
		
		y = 32
		screen.blit(title, (32, y))
		y += 16
		index = 0
		for option in self.options:
			selected = False
			if index == self.index:
				screen.blit(cursor, (left, y))
				selected = True
			
			if not selected or self.blink_counter <= 0 or rc % 4 < 2:
				screen.blit(getText((255, 255, 255), option), (left + 16, y))
			index += 1
			y += 16
		
		

SONG_LOOKUP = {
	# name : (file, should loop?)
	'title': ('tensebranches', True),
	'intro': ('open_skies', True),
	'overworld': ('open_skies', True),
	'ship': ('spacemachine', True),
	'credits': ('credits', True),
	'password': ('fluffytales', True),
	'volcano': ('lunacy', True),
	'bike': ('nopressure', True),
	'water': ('squirrels_can_swim', True),
	'gravity': ('deadlyacorns', True),
	'death': ('oopsdead', False),
	'gameover': ('fluffytales', True),
}

SONG_BY_MAP = {
	'ship_1': 'ship',
	
	'main': 'overworld',
	'balloon_2_overland': 'overworld',
	'balloon_3_overland': 'overworld',
	'bird_overland': 'overworld',
	'volcano_2_overland': 'overworld',
	'volcano_3_overland': 'overworld',
	
	'bike_level': 'bike',
	'bird_link': 'gravity',
	
	'exit_water_phase_locked': 'water',
	'exit_water_phase_unlocked': 'water',
	
	'grav_ascent': 'gravity',
	'grav_core': 'gravity',
	'grav_descent': 'gravity',
	'grav_descent_secret': 'gravity',
	'grav_gate': 'gravity',
	
	'lava_gate_2': 'volcano',
	'lava_gate_2_active': 'volcano',
	'lava_gate_3': 'volcano',
	'lava_gate_3_active': 'volcano',
	'volcano_1_ascent': 'volcano',
	'volcano_1_core': 'volcano',
	'volcano_1_descent': 'volcano',
	'volcano_2_core': 'volcano',
	'volcano_2_lower_ascent': 'volcano',
	'volcano_2_lower_descent': 'volcano',
	'volcano_2_upper_ascent': 'volcano',
	'volcano_2_upper_descent': 'volcano',
	'volcano_2_upper_secret': 'volcano',
	'volcano_3_bottom': 'volcano',
	'volcano_3_core': 'volcano',
	'volcano_3_middle': 'volcano',
	'volcano_3_secret': 'volcano',
	'volcano_3_top': 'volcano',
	
	'water_gate_1_locked': 'water',
	'water_gate_1_unlocked': 'water',
	
	'rope_test': 'gravity',
	
	'birdfeeder': 'gravity',
	
	'section_3_gate': 'gravity',
	'section_3_gate_active': 'gravity',
	'wlink_A1': 'gravity',
	'wlink_A2': 'gravity',
	'wlink_A3': 'gravity',
	'wlink_A4': 'gravity',
	
	'wlink_B_nowater': 'water',
	'wlink_B_somewater': 'water',
	'wlink_B_allwater': 'water',
	
	'wlink_D1_nowater': 'water',
	'wlink_D1_water': 'water',
	'wlink_D2_nowater': 'water',
	'wlink_D2_water': 'water',
	'wlink_D3_nowater': 'water',
	'wlink_D3_water': 'water',
	
	'wlink_E1_nowater': 'water',
	'wlink_E1_water': 'water',
	'wlink_E2': 'water',
	'wlink_E3': 'water',
}

for file in os.listdir('maps'):
	if file.endswith('.map'):
		if 'overlay' in file:
			pass
		else:
			song = SONG_BY_MAP.get(file.split('.')[0])
			if song == None:
				print("ERROR: no song assigned to map file: " + file)



class JukeBox:
	def __init__(self):
		self.currentSong = None
		self.actuallyPlaying = False
	
	def ensureSong(self, id):
		if not SOUND_ENABLED:
			return
		
		if self.currentSong != id:
			if self.currentSong != None and self.actuallyPlaying:
				pygame.mixer.music.stop()
			song = SONG_LOOKUP.get(id)
			if song != None:
				file = song[0]
				loop = song[1]
				path = 'music/' + song[0] + '.ogg'
				pygame.mixer.music.load(path.replace('/', os.sep))
				pygame.mixer.music.play(-1 if loop else 0)
				self.actuallyPlaying = True
			else:
				self.actuallyPlaying = False
			
			self.currentSong = id
	
	def playSongForLevelMaybe(self, id):
		song = SONG_BY_MAP.get(id)
		if song != None:
			self.ensureSong(song)
	
JUKEBOX = JukeBox()

class MapParser:
	def __init__(self, file):
		path = ('maps/' + file).replace('/', os.sep).replace('\\', os.sep)
		c = open(path, 'rt')
		t = c.read()
		c.close()
		values = {}
		for line in t.split('\n'):
			parts = line.split(':')
			if len(parts) < 2:
				continue
			
			key = trim(parts[0][1:])
			value = trim(':'.join(parts[1:]))
			values[key] = trim(value)
		self.values = values
	
	def parse(self):
		m = StemCell()
		m.width = int(self.values['width'])
		m.height = int(self.values['height'])
		m.side = self.values.get('view', 'side').lower() == 'side'
		m.upper = self.getTiles(self.values['upper'], m.width, m.height)
		m.lower = self.getTiles(self.values['lower'], m.width, m.height)
		m.doors = self.getDoors(self.values.get('doors', ''))
		m.enemies = self.getEnemies(self.values.get('enemies', ''))
		m.doorswaps = self.getDoorswaps(self.values.get('doorswaps', ''))
		m.overlayTriggers = self.getOverlayTriggers(self.values.get('overlay', ''))
		m.powerups = self.getPowerups(self.values.get('powerups', ''))
		return m
	
	def getPowerups(self, sv):
		sv = trim(sv)
		if len(sv) == 0:
			return []

		output = []
		for item in sv.split(','):
			parts = item.split('|')
			pu = StemCell()
			pu.type = parts[0]
			pu.id = parts[1]
			pu.col = int(parts[2])
			pu.row = int(parts[3])
			output.append(pu)
		return output
		
	def getOverlayTriggers(self, strValue):
		strValue = trim(strValue)
		if len(strValue) == 0:
			return {}
		
		output = {}
		items = strValue.split(',')
		for item in items:
			parts = item.split('|')
		
			if len(parts) == 2:
				output[trim(parts[0])] = trim(parts[1])
		
		return output
	
	# door swap output format:
	# { original ID => List[ Pair<trigger, swapped ID> ] }
	def getDoorswaps(self, strValue):
		strValue = trim(strValue)
		if len(strValue) == 0:
			return {}
		swaps = {}
		for swap in strValue.split(','):
			parts = swap.split('|')
			trigger = trim(parts[0])
			original = trim(parts[1])
			swapped = trim(parts[2])
			
			swaps[original] = swaps.get(original, [])
			swaps[original].append((trigger, swapped))
		return swaps
		
	
	def getEnemies(self, enemyString):
		enemyString = trim(enemyString)
		if len(enemyString) == 0:
			return []
		enemies = []
		for value in enemyString.split(','):
			parts = value.split('|')
			id = parts[0]
			col = int(parts[1])
			row = int(parts[2])
			e = StemCell()
			e.id = id
			e.col = col
			e.row = row
			enemies.append(e)
		return enemies
	
	def getDoors(self, doorString):
		doorString = trim(doorString)
		if len(doorString) == 0:
			return []
		doors = []
		for door in doorString.split(','):
			door = door.split('|')
			d = StemCell()
			d.target = door[0]
			d.sx = int(door[1])
			d.sy = int(door[2])
			d.tx = int(door[3])
			d.ty = int(door[4])
			doors.append(d)
		return doors
		
	def getTiles(self, ids, width, height):
		output = []
		for id in map(trim, ids.split(',')):
			if id == '':
				output.append(None)
			else:
				output.append(getTile(id))
		
		grid = makeGrid(width, height)
		y = 0
		index = 0
		while y < height:
			x = 0
			while x < width:
				grid[x][y] = output[index]
				index += 1
				x += 1
			y += 1
		return grid

class MyEvent:
	def __init__(self, action, down):
		self.action = action
		self.down = down
		self.up = not down
	


class OpeningScene:
	def __init__(self):
		self.static_width = 150
		if GO_TO_TITLE:
			self.next = TitleScene()
			self.mode = None
			self.flags = ''
			self.fade = 0
			self.mode_counter = 0
		else:
			self.mode = 'open'
			self.ch3 = [
				' XXX  X  X   XXX ',
				'XX XX X  X  X   X',
				'X     X  X      X',
				'X     XXXX    XX ',
				'X     X  X      X',
				'XX XX X  X  X   X',
				' XXX  X  X   XXX '
				]
			self.logo = [
				'TTTTT TTTTT  TTT  T     T       N   N NNNN    NNNN  NNNN  NNNNN  NNN  NNNNN N   N NNNNN  NNN ',
				'  T   T     T   T TT   TT       NN  N N   N   N   N N   N N     N   N N     NN  N   N   N   N',
				'  T   T     T   T T T T T       N N N N   N   N   N N   N N     N     N     N N N   N   N    ',
				'  T   TTT   TTTTT T  T  T       N  NN NNNN    NNNN  NNNN  NNN    NNN  NNN   N  NN   N    NNN ',
				'  T   T     T   T T     T       N   N N       N     N N   N         N N     N   N   N       N',
				'  T   T     T   T T     T       N   N N       N     N  N  N     N   N N     N   N   N   N   N',
				'  T   TTTTT T   T T     T       N   N N       N     N   N NNNNN  NNN  NNNNN N   N   N    NNN ',
			]
			self.next = self
			self.flags = 'W'
			static = pygame.Surface((self.static_width, self.static_width))
			self.scrolly = pygame.Surface((WIDTH, 5))
			black = (0, 0, 0)
			white = (255, 255, 255)
			self.dark = pygame.Surface((WIDTH, HEIGHT)).convert()
			self.dark.fill((0, 0, 0))
			
			dr = pygame.draw.rect
			pr = pygame.Rect
			e = 0
			y = 0
			self.fade = 0
			colors = [black, white] * (self.static_width * self.static_width // 2)
			random.shuffle(colors)
			i = 0
			while y < self.static_width:
				x = 0
				while x < self.static_width:
					dr(static, colors[i], pr(x, y, 2, 2))
					i += 1
					x += 1
				y += 1
			self.static = static
			self.mode_counter = 0
		
	def processInput(self, events, pressedKeys):
		for event in events:
			
			if self.mode == 'open':
				if event.down and event.action in ('start', 'A', 'B'):
					self.mode = 'starting'
					self.mode_counter = 0
			
	def update(self):
		self.fade += 1
		self.mode_counter += 1
		if self.mode == 'starting' and self.mode_counter > 20:
			self.next = TitleScene()
		if self.mode == 'open':
			if self.mode_counter > 200:
				self.mode = 'starting'
				self.mode_counter = 0
	
	def render(self, screen, renderCounter):
		if self.mode == 'starting':
			if self.mode_counter < 10:
				screen.fill((255, 80, 255))
			else:
				screen.fill((200, 200, 200))
		elif self.mode == 'open':
			left = -random.random() * self.static_width
			top = -random.random() * self.static_width
			y = top
			while y < HEIGHT:
				x = left
				while x < WIDTH:
					screen.blit(self.static, (x, y))
					x += self.static_width
				y += self.static_width
			
			self.scrolly.blit(screen, (0, 0))
			
			for i in range(5):
				y = (renderCounter * 2 + 10 * i + 300) % HEIGHT
				screen.blit(self.scrolly, (0, y))
				
			width = len(self.ch3[0])
			height = len(self.ch3)
			
			y = 0
			while y < height:
				x = 0
				while x < width:
					if self.ch3[y][x] != ' ':
						pygame.draw.rect(screen, (0, 255, 0), pygame.Rect(20 + x * 8, 20 + y * 8, 8, 8))
					x += 1
				y += 1
			
			width = len(self.logo[0])
			height = len(self.logo)
			y = 0
			while y < height:
				x = 0
				while x < width:
					if self.logo[y][x] != ' ':
						pygame.draw.rect(screen, (0, 255, 0), pygame.Rect(20 + x * 8, 250 + y * 8, 8, 8))
					x += 1
				y += 1
			opacity = int(self.fade * 1.5)
			if opacity > 255:
				opacity = 255
			
			v = 255 - opacity
			self.dark.set_alpha(v)
			if (v > 0):
				screen.blit(self.dark, (0, 0))
				

PASSWORD_BLOCK = [
	'123456789',
	'0ABCDEFGH',
	'IJKLMNOPQ',
	'RSTUVWXYZ']

class PasswordScene:
	def __init__(self):
		self.next = self
		self.flags = ''
		self.row = 0
		self.col = 0
		self.blinkCounter = 0
		self.current = ''
	
	def processInput(self, events, pressed):
		width = len(PASSWORD_BLOCK[0])
		height = len(PASSWORD_BLOCK)
		
		for event in events:
			dx = 0
			dy = 0
			if event.down:
				if event.action == 'up':
					dy -= 1
				elif event.action == 'down':
					dy += 1
				elif event.action == 'left':
					dx -= 1
				elif event.action == 'right':
					dx += 1
				elif event.action in ('A', 'start'):
					if self.row == 4 and self.col >= 5:
						self.next = TitleScene()
					else:
						if self.row == 4:
							self.tryCommit()
						else:
							if len(self.current) == 4:
								playNoise('head_bonk')
							else:
								if self.row < 4:
									if len(self.current) == 4:
										playNoise('head_bonk')
									else:
										playNoise('password_enter_digit')
										self.current += PASSWORD_BLOCK[self.row][self.col]
								
				elif event.action == 'B':
					if len(self.current) > 0:
						self.current = self.current[:-1]
					else:
						playNoise('head_bonk')
			
			newCol = self.col + dx
			newRow = self.row + dy
			if dx != 0 or dy != 0:
				bonk = False
				if self.row == 4:
					if dy == -1:
						pass #allow
					elif dy == 1:
						bonk = True
					elif dx == -1:
						if self.col >= 5:
							newCol = 0
						else:
							bonk = True
					elif dx == 1:
						if self.col < 5:
							newCol = 6
						else:
							self.bonk = True
				else:
					if newCol < 0 or newRow < 0:
						self.bonk = True
					if newCol >= width:
						self.bonk = True
				if not bonk:
					self.col = newCol
					self.row = newRow
					playNoise('menu_move')
				else:
					playNoise('head_bonk')
	
	def tryCommit(self):
		if len(self.current) != 4:
			playNoise('bad_password')
			return
		for i in (0, 1, 3):
			if not (self.current[i] in ALPHABET):
				playNoise('bad_password')
				return
		
		if not (self.current[2] in '0123456789'):
			playNoise('bad_password')
			return
		
		context = Context()
		context.convertFromPassword(self.current)
		
		self.next = PlayScene('ship_1', 8, 9, context)
		
				
	
	def update(self):
		self.blinkCounter += 1
		JUKEBOX.ensureSong('password')

	def getCoord(self, col, row):
		return (8 + col * 24 + 24, 64 + row * 16)
	
	def render(self, screen, rc):
		screen.fill((0, 0, 0))
		
		hleft = 48
		drawBox(screen, hleft, 0, 18, 5)
		header = getText(WHITE, "Password Entry")
		screen.blit(header, (hleft + 16, 16))
		
		gleft = 8
		gtop = 48
		drawBox(screen, gleft, gtop, 30, 13)
		pb = PASSWORD_BLOCK
		blink = self.blinkCounter % 16 < 8
		height = len(pb)
		width = len(pb[0])
		y = 0
		while y < height:
			x = 0
			while x < width:
				letter = getText(WHITE, pb[y][x])
				c = self.getCoord(x, y)
				screen.blit(letter, c) # (gleft + 24 + x * 24, gtop + 16 + y * 16)
				if self.col == x and self.row == y:
					if blink:
						screen.blit(getText(WHITE, '>'), (c[0] - 8, c[1]))
				x += 1
			y += 1
		
		x = 0
		y = 4
		px = gleft + 24 + x * 24
		py = gtop + 16 + y * 16
		screen.blit(getText(WHITE, "DONE"), (px, py))
		if self.row == 4 and self.col < 5 and blink:
			screen.blit(getText(WHITE, '>'), (px - 8, py))
		x = 6
		px = gleft + 24 + x * 24
		screen.blit(getText(WHITE, "CANCEL"), (px, py))
		if self.row == 4 and self.col >= 5 and blink:
			screen.blit(getText(WHITE, '>'), (px - 8, py))
		
		oleft = 88
		otop = 224 - 64
		drawBox(screen, oleft, otop, 8, 5)
		txt = getText(WHITE, self.current)
		screen.blit(txt, (oleft + 16, otop + 16))

class PauseScene:
	def __init__(self, bg):
		self.bg = bg
		self.next = self
		self.flags = ''
		self.counter = 0
	
	def processInput(self, events, pressed):
		for event in events:
			if event.down and event.action == 'start':
				if 'M' in self.bg.flags:
					playNoise('pause_sound')
					pygame.mixer.music.unpause()
				self.next = self.bg
				self.bg.next = self.bg
	
	def update(self):
		if 'M' in self.bg.flags:
			if self.counter == 0:
				pygame.mixer.music.pause()
				playNoise('pause_sound')
		
		self.counter += 1
	
	def render(self, screen, rc):
		screen.fill(BLACK)
		img = getText(WHITE, "PAUSE")
		x = 128 - img.get_width() // 2
		y = 112 - img.get_height() // 2
		screen.blit(img, (x, y))

WHITE = (255, 255, 255)
GOOD_HEALTH_COLOR = WHITE
POOR_HEALTH_COLOR = (255, 128, 128)

BIKE_SPEED = 6

class PlayScene:
	def __init__(self, map, startCol, startRow, context):
		self.startArgs = (map, startCol, startRow)
		self.updateCounter = 0
		self.context = context
		self.next = self
		self.flags = 'M'
		levelname = map.split('.')[0]
		if DEBUG_MODE:
			print("Active map:", levelname)
		self.id = levelname
		if not map.endswith('.map'):
			map += '.map'
		mapParser = MapParser(map)
		map = mapParser.parse()
		
		self.hasAtmosphere = context.volcanoA and context.volcanoB and context.volcanoC

		self.bg = getBackground(levelname)
		stars = legacyMap(lambda x:getImage('tiles/background/stars' + str(x) + '.png'), [1, 2, 3, 4, 5]) + ([None] * 10)
		stars *= 3
		random.shuffle(stars)
		self.stars = stars
		self.runCounter = 0
		self.runCounterValidFor = ':P'
		
		self.cols = map.width
		self.rows = map.height
		self.upper = map.upper
		self.lower = map.lower
		self.side = map.side
		
		self.lazor_cooldown = 0
		
		self.tiles = makeGrid(self.cols, self.rows)
		
		self.doorSwaps = map.doorswaps
		
		doorTiles = []
		ladderTiles = {}
		lowerdoors = []
		y = self.rows - 1
		while y >= 0:
			x = 0
			while x < self.cols:
				
				t = makeTile(self.lower[x][y], self.upper[x][y], x, y)
				self.tiles[x][y] = t
				if t.isDoor:
					doorTiles.append((str(x) + '|' + str(y), t))
					if self.side and y + 1 < self.rows:
						# add the tile below the door to the potential door tiles
						# the door list from the map parser will check for the tile below a door iff it's a tall door and is 
						# expecting to find it in this list. If not, then this gets safely ignored
						doorTiles.append((str(x) + '|' + str(y + 1), self.tiles[x][y + 1]))
						
				if t.isLadder:
					ladderTiles[str(x) + '|' + str(y)] = (x, y)
				x += 1
			y -= 1 # go backwards so doors can add lower neighbor
		
		for lk in ladderTiles.keys():
			coord = ladderTiles[lk]
			x = coord[0]
			y = coord[1]
			if y > 0:
				if ladderTiles.get(str(x) + "|" + str(y - 1)) == None:
					self.tiles[x][y].isTop = True
		
		doorLookup = {}
		for door in map.doors:
			xys = [(door.sx, door.sy)]
			if self.side:
				xys.append((door.sx, door.sy + 1))
			for xy in xys:
				doorLookup[str(xy[0]) + '|' + str(xy[1])] = door
		
		for doorTile in doorTiles:
			dk = doorTile[0]
			door = doorLookup.get(dk, None)
			if door != None:
				doorTile[1].door = door
				doorTile[1].collisions = []
				doorTile[1].solid = False
		
		self.cameraX = 0
		self.cameraY = 0
		
		self.player = Sprite('player_' + ('side' if self.side else 'over'), startCol * 16 + 8, startRow * 16 + 7)
		self.sprites = [self.player]
		
		if levelname == 'bike_level':
			self.player.bikemode = True
			self.player.floats = True
		
		for enemy in map.enemies:
			sprite = Sprite(enemy.id, enemy.col * 16 + 8, enemy.row * 16 + 8)
			sprite.isEnemy = True
			self.sprites.append(sprite)
		
		self.overlayTriggers = map.overlayTriggers
		
		for powerup in map.powerups:
			puid = powerup.id
			if context.powerupsTaken.get(puid, False):
				pass
			else:
				powerupSprite = Sprite('powerup', powerup.col * 16 + 8, powerup.row * 16 + 7)
				powerupSprite.powerupInfo = powerup
				self.sprites.append(powerupSprite)
		
		self.special = getSpecialLevelStuff(levelname, self)
		for special in self.special:
			if special.hasPostInit:
				special.postInit()
	
	def playersTile(self, offsetX=0, offsetY=0):
		if self.player == None: return None
		p = self.player
		tx = int(p.modelX / 16) + offsetX
		ty = int(p.modelY / 16) + offsetY
		if tx >= 0 and ty >= 0 and tx < self.cols and ty < self.rows:
			return self.tiles[tx][ty]
		return None
	
	def processInput(self, events, pressed):
		frozen = False
		for s in self.special:
			if s.freeze:
				frozen = True
		
		self.shooting = self.lazor_cooldown > 0
		
		pause = False
		
		if not frozen:
			if self.side:
				dx = 0
				dy = 0
				running = False
				if self.context.gravity:
					if pressed['B'] and not self.player.cling:
						running = True
						v = 5
					else:
						v = 3
				else:
					v = 2.5
				
				if self.player.bikemode:
					v = 2.0
				
				
				if pressed['left']:
					dx = -v
					if running and self.runCounterValidFor == 'left':
						self.runCounter += 1
					else:
						self.runCounter = 0
						self.runCounterValidFor = 'left'
				elif pressed['right']:
					dx = v
					if running and self.runCounterValidFor == 'right':
						self.runCounter += 1
					else:
						self.runCounter = 0
						self.runCounterValidFor = 'right'
				else:
					self.runCounter = 0
					self.runCounterValidFor = 'nothing'
					
				
				if pressed['up']:
					
					if self.player.bikemode:
						dy = -v
					
					pt = self.playersTile()
					if pt != None and pt.isLadder:
						self.player.cling = True
						#self.player.onGround = False
						self.player.ladderDY = -2
				elif pressed['down']:
					
					if self.player.bikemode:
						dy = v
					
					pt = self.playersTile()
					if pt != None and pt.isLadder:
						self.player.ladderDY = 2
					else:
						pt = self.playersTile(0, 1)
						if pt != None and pt.isLadder:
							self.player.modelY += 8
							self.player.ladderDY = 2
							self.player.cling = True
				
				if self.player != None:
					if self.player.bikemode:
						dx += BIKE_SPEED
					self.player.dx = dx
					self.player.dy = dy
				
				for event in events:
					if event.action == 'A':
						if event.down:
							pt = self.playersTile()
							if self.player.onGround or self.player.cling or (pt != None and pt.isWater):
								self.player.onGround = False
								self.player.cling = False
								self.player.ladderDY = 0
								if pt.isWater:
									self.player.vy += WATER_JUMPING_VY
									playNoise('swim')
								else:
									self.player.vy = JUMPING_VY
									if self.runCounter > 5 and self.context.gravity:
										self.player.vy = RUN_JUMPING_VY
									playNoise('jump')
						else:
							if self.player.vy < 0:
								self.player.vy = self.player.vy / 4.0 # maybe set to 0 instead?
					
					# SHOOT UR LAZOR PEW PEW!!!!1
					elif event.action == 'B' and event.down:
						if self.lazor_cooldown <= 0:#and not self.player.cling:
							p = self.player
							lazorLeft = self.player.lastDirection == 'left'
							if self.player.damageDir != None:
								lazorLeft = self.player.damageDir == 'left'
							lazorVX = 8
							if lazorLeft:
								lazorVX = -8
							
							x = p.x + lazorVX
							y = p.y - 8
							
							if p.spawns == None:
								p.spawns = []
							
							playNoise('lazor')
							bullet = Sprite('lazor', x, y)
							bullet.bvx = lazorVX
							bullet.isBullet = True
							bullet.ghost = True
							bullet.floats = True
							p.spawns.append(bullet)
							self.lazor_cooldown = 6
					elif event.action == 'start' and event.down:
						self.next = PauseScene(self)
			else:
				v = 3
				dx = 0
				dy = 0
				
				if pressed['left']:
					dx = -v
				elif pressed['right']:
					dx = v
					
				if pressed['up']:
					dy = -v
				elif pressed['down']:
					dy = v
				
				for event in events:
					if event.action == 'start' and event.down:
						self.next = PauseScene(self)
				
				self.player.dx = dx
				self.player.dy = dy
	
	def playMusic(self):
		JUKEBOX.playSongForLevelMaybe(self.id)
	
	def update(self):
		self.updateCounter += 1
		self.playMusic()
		
		playerX = self.player.modelX
		playerY = self.player.modelY
		sprites = self.sprites
		self.lazor_cooldown -= 1
		newsprites = [] # TODO: filter out the dead
		
		allBullets = []
		i = 0
		while i < len(sprites):
			sprite = sprites[i]
			if sprite != None and not sprite.dead:
				sprite.update(self)	
				if sprite.isEnemy and self.player != None:
					if sprite.isCollision(self.player):
						self.playerHit()
				
				if sprite.spawns != None:
					for spawn in sprite.spawns:
						sprites.append(spawn)
				sprite.spawns = None
				
				if sprite.isBullet:
					allBullets.append(sprite)
			i += 1
		
		new_sprites = []
		
		i = 0
		while i < len(sprites):
			sprite = sprites[i]
			if len(allBullets) > 0 and sprite.type != 'bird1' and sprite.type != 'bird2':
				if sprite.isEnemy:
					left = sprite.x - 8
					right = sprite.x + 8
					bottom = sprite.y + 8
					top = bottom - sprite.height
					
					j = 0
					while j < len(allBullets):
						bullet = allBullets[j]
						bleft = bullet.x - 8
						bright = bleft + 16
						bbottom = bullet.y + 4
						btop = bbottom - 8
						
						if bleft > right:
							pass
						elif bright < left:
							pass
						elif btop > bottom:
							pass
						elif bbottom < top:
							pass
						else:
							sprite.dead = True
							bullet.dead = True
							if sprite.type in ('acorn', 'acorntop', 'byat', 'moonsquid', 'lavamonster'):
								dropN = int(random.random() * 200)
								drop = None
								if dropN == 0:
									drop = '1up'
								elif dropN < 8:
									drop = 'life_big'
								elif dropN < 30:
									drop = 'life_small'
								
								if drop != None:
									# Can't do spawns since it'll get filtered out before then.
									powerup = Sprite('powerup', sprite.x, sprite.y - 1)
									info = StemCell()
									info.id = None
									info.type = drop
									powerup.powerupInfo = info
									new_sprites.append(powerup)
							
							poofType = 'poof_small'
							if sprite.type == 'lavamonster':
								poofType = 'poof_big'
							poof = Sprite(poofType, sprite.x, sprite.y)
							new_sprites.append(poof)
							playNoise('enemy_dies')
						
						j += 1
			
			if not sprite.dead:
				new_sprites.append(sprite)
			i += 1
		
		self.sprites = new_sprites
		
		player_tx = int(self.player.modelX / 16)
		player_ty = int(self.player.modelY / 16)
		activeTile = self.tiles[player_tx][player_ty]
		if activeTile.door != None:
			door = activeTile.door
			target = door.target
			swaps = self.doorSwaps.get(target)
			
			if swaps != None:
				ctx = self.context
				for swap in swaps:
					trigger = swap[0]
					if ((
						trigger == 'gravity' and ctx.gravity) or (
						trigger == 'lavaA' and ctx.volcanoA) or (
						trigger == 'lavaB' and ctx.volcanoB) or (
						trigger == 'lavaC' and ctx.volcanoC) or (
						trigger == 'waterA' and ctx.balloonA) or (
						trigger == 'waterB' and ctx.balloonB) or (
						trigger == 'waterC' and ctx.balloonC) or (
						trigger == 'waterD' and ctx.balloonD)):
						target = swap[1]
						break
			
			for special in self.special:
				if special.hasDoorTrigger:
					target = special.doorTrigger(target)
					break
			
			if target != None:
				self.next = PlayScene(target, door.tx, door.ty, self.context)
		
		pUp = self.playersTile(0, -1)
		pDown = self.playersTile(0, 0)
		
		if (pUp != None and pUp.isSpike) or (pDown != None and pDown.isSpike):
			self.playerHit(3)
		
		if (pUp != None and pUp.isLava) or (pDown != None and pDown.isLava):
			self.next = DeathOverrideScene(self, 'lava')
		
		for special in self.special:
			if special.hasUpdate:
				special.update()
	
	def playerHit(self, amount=1):
		if self.player.blinkCounter < 0:
			self.player.hit(self, amount)
	
	def isCollision(self, pLeft, pTop, pRight, pBottom):
		if pLeft < 0: return True
		if pTop < 0: return True
		tLeft = int(pLeft / 16)
		tRight = tLeft if (pRight == pLeft) else int(pRight / 16)
		tTop = int(pTop / 16)
		
		# potentially a bug
		# bottom row of sprite is technically top row of ground below. This intersection should be ignored.
		tBottom = int((pBottom - 3) / 16) 
		
		if tLeft < 0: return True
		if tTop < 0: return True
		if tRight >= self.cols: return True
		if tBottom >= self.rows: return True
		
		y = tTop
		while y <= tBottom:
			x = tLeft
			while x <= tRight:
				if self.tiles[x][y].solid:
					return True
				x += 1
			y += 1
		return False
	
	def renderOverlay(self, screen):
		pygame.draw.rect(screen, (0, 0, 0), pygame.Rect(0, 0, 256, 8))
		
		screen.blit(getImage('misc/lives.png'), (0, 0))
		txt = getText(WHITE, 'x' + str(self.context.lives))
		screen.blit(txt, (8, 0))
		
		x = 112
		txt = getText(WHITE, 'HEALTH')
		screen.blit(txt, (x, 0))
		
		x += txt.get_width() + 8
		
		health = self.context.lifemeter
		color = GOOD_HEALTH_COLOR
		if health < 4:
			color = POOR_HEALTH_COLOR
		i = 0
		while i < 5:
			pygame.draw.rect(screen, (128, 128, 128), pygame.Rect(x, 0, 16, 8), 1)
			if health >= ((i + 1) * 2):
				pygame.draw.rect(screen, color, pygame.Rect(x + 1, 1, 14, 6))
			elif health == 2 * i + 1:
				pygame.draw.rect(screen, color, pygame.Rect(x + 1, 1, 7, 6))
			
			x += 16
			i += 1
		
		cat = self.context.getCurrentCountAndType()
		if cat != None:
			cnt = cat[1]
			
			if cat[0] == 'W':
				img = getImage('misc/water_icon.png')
				total = 4
			else:
				img = getImage('misc/volcano_icon.png')
				total = 3
							
			screen.blit(img, (48, 0))
			screen.blit(getText(WHITE, str(cnt) + '/' + str(total)), (56, 0))
	def render(self, screen, rc):
		if self.bg == 'stars':
			screen.fill((0, 0, 0))
		elif self.bg == 'cave':
			pass
		elif self.bg == 'volcano':
			pass
		else:
			screen.fill((0, 0, 40)) # sky
		
		colStart = 0
		colEnd = self.cols - 1
		
		rowStart = 0
		rowEnd = self.rows - 1
		
		offsetX = -(self.player.x - 128)
		offsetY = -(self.player.y - 112)
		
		if self.player.bikemode:
			offsetX = -(self.updateCounter * BIKE_SPEED)
		
		if offsetX > 0: offsetX = 0
		if offsetY > 0: offsetY = 0
		
		right = -(self.cols * 16 - 256)
		if offsetX < right: offsetX = right
		
		bottom = -(self.rows * 16 - 224)
		if offsetY < bottom: offsetY = bottom
		
		if self.cols * 16 < 256:
			offsetX = (256 - self.cols * 16) // 2
		if self.rows * 16 < 224:
			offsetY = (224 - self.rows * 16) // 2
		
		cave = getImage('tiles/background/cave.png')
		volcano_bg = getImage('tiles/background/volcano.png')
		starlen = len(self.stars)
		stars = self.stars
		counter = 0
		
		idealColStart = -int(offsetX / 16) - 1
		if idealColStart > colStart:
			colStart = idealColStart
		
		idealColEnd = colStart + 17
		if idealColEnd < colEnd:
			colEnd = idealColEnd
		
		idealRowStart = -int(offsetY / 16) - 1
		if idealRowStart > rowStart:
			rowStart = idealRowStart
		
		idealRowEnd = rowStart + 15
		if idealRowEnd < rowEnd:
			rowEnd = idealRowEnd
		
		lavaLevel = 0
		shake = 0
		for s in self.special:
			lavaLevel = max(lavaLevel, s.lavaLevel)
			shake = max(shake, 3 if s.shakeScreen else 0)
		
		offsetX += shake
		offsetY += shake // 3
		
		lavaRowType = None
		
		row = rowStart
		while row <= rowEnd:
			
			fromBottom = rowEnd - row + 1
			if lavaLevel > 0:
				if fromBottom == lavaLevel:
					lavaRowType = 'upper'
					lc = str(int((rc // 4) % 4) + 1)
					lu = getImage('tiles/fluids/lava' + lc + '.png')
				elif fromBottom < lavaLevel:
					lavaRowType = 'inner' 
					lu = getImage('tiles/fluids/lava.png')
			
			
			
			col = colStart
			while col <= colEnd:
				counter += 1
				x = col * 16 + offsetX
				y = row * 16 + offsetY
				pt = (x, y)
				if self.bg == 'cave':
					screen.blit(cave, pt)
				elif self.bg == 'volcano':
					screen.blit(volcano_bg, pt)
					if lavaRowType != None:
						screen.blit(lu, pt)
					
				elif self.bg == 'stars':
					bgimg = stars[(col + self.rows * row + row * row) % starlen]
					if bgimg != None:
						screen.blit(bgimg, pt)
				templates = self.tiles[col][row].templates
				i = 0
				while i < len(templates):
					img = templates[i].getImage(rc)
					screen.blit(img, pt)
					i += 1
				
				col += 1
			row += 1
		
		re_render = []
		arc = rc // 4
		for sprite in self.sprites:
			if sprite != None and not sprite.dead:
				sprite.render(self, screen, offsetX, offsetY, arc)
				if sprite.inBackground:
					tx = sprite.x // 16
					ty = (sprite.y + 8) // 16
					re_render.append((tx, ty))
					re_render.append((tx, ty - 1))
					re_render.append((tx, ty - 2))
		
		for rr in re_render:
			col = rr[0]
			row = rr[1]
			if row >= 0:
				x = col * 16 + offsetX
				y = row * 16 + offsetY
				pt = (x, y)
				templates = self.tiles[col][row].templates
				i = 0
				while i < len(templates):
					img = templates[i].getImage(rc)
					screen.blit(img, pt)
					i += 1
				
		
		for s in self.special:
			if s.hasRender:
				s.render(screen, rc, offsetX, offsetY)
		
				
		self.renderOverlay(screen)
	
	def triggerDialog(self, id, slightDelay=False):
		self.next = DialogScene(self, id, slightDelay)
		

class ReadyScene:
	def __init__(self, previousScene, context):
		self.next = self
		self.flags = ''
		self.context = context
		self.startArgs = previousScene.startArgs
		self.counter = 0
		self.playscene = PlayScene(self.startArgs[0], self.startArgs[1], self.startArgs[2], self.context)
	
	def processInput(self, events, pressed):
		pass
	
	def update(self):
		self.playscene.playMusic()
		self.counter += 1
		if self.counter > 90:
			self.next = self.playscene
	
	def render(self, screen, rc):
		if self.counter < 21:
			screen.fill((0, 0, 0))
		else:
			self.playscene.render(screen, 0)
			txt = getText(WHITE, 'ready')
			if (rc % 8 < 4):
				screen.blit(txt, (112, 112))

# This maps sound IDs used in the code to actual file names.
# If entry is None, no sound will be played when that event is triggered
# Otherwise, entry is a tuple of the filename (without extension, sfx/ is assumed) followed by a normalization ratio.

SOUND_LOOKUP = {
	'head_bonk': ('head_bonk',1.0), # fairly sublte sound as it will also happen a lot. Re-used when trying to move a menu cursor someplace where it cannot go.
	'password_enter_digit':('password_digit_enter',1.0), 
	'bad_password': ('bad_password',1.0), # sad fanfare: "DOO DOOoooo"
	'sprinkle_lava_packet': ('sprinkle_lava',1.0), # "chkhchkhchkchkh" no longer than 1 second
	'screen_shaking': ('everythingfalls', 1.0), # done. perfect.
	'fall_to_death': ('fall_to_death',1.0), # fairly obvious
	'lava_rise': ('crash', 1.0),
	'land_on_ground': ('land_on_ground',0.5), # subtle sound as it will happen a lot
	'water_pop': ('crash', 1.0),
	
	# TODO(Blake): make sure these are hooked up in all menus
	'menu_select': ('menuhigh', 6.0), # positive sound
	'menu_move': ('menulow', 4.0), # subtle
	
	'low_health': ('low_health',1.0), # annoying as hell, 5 high pitched beeps in quick succession
	'get_hit': ('get_hit',1.0), # use megaman taking-damage sound as inspiration, if that's doable.
	'raise_health': ('raise_health', 1.0), # dwoooOOO (but short)
	'1up': ('1up',1.0),
	'jump': ('jump', 0.4), 
	'swim': ('jump', 0.4), # maybe use a different sound?
	'pause_sound': ('pause_sound',1.0), # "TEE KOO TEE KOO!"
	'lava_roast': ('lava_roast',1.0), # player dies by lava
	'text_char': ('text_char', 0.3), # subtle text character bloop
	'lazor': ('lazor',1.0), # player shoots lazor
	
	# TODO(Blake): The following are not hooked in by the code yet
	'enemy_dies': ('enemy_dies',1.0), # after shooting it
}

_sound_cache = {}
def playNoise(key):
	if not SOUND_ENABLED:
		return
	
	snd = _sound_cache.get(key)
	if snd == None:
		data = SOUND_LOOKUP.get(key)
		if data != None:
			filename = data[0]
			volumeRatio = data[1]
			path = 'sfx/' + filename + '.ogg'
			snd = pygame.mixer.Sound(path.replace('/', os.sep))
			snd.set_volume(volumeRatio)
			_sound_cache[key] = snd
	
	
	if snd != None:
		snd.play()
	


class BalloonPopping(SpecialLevelStuff):
	# ID's: water1, water2, water3, water4, lava1, lava2, lava3
	def __init__(self, scene, id):
		SpecialLevelStuff.__init__(self, scene)
		self.id = id
		self.scene = scene
		self.context = scene.context
		self.hasUpdate = True
		self.hasPostInit = True
		self.done = False
		
		self.balloonCoords = {}
		
		if id.startswith('water'):
			rows = scene.rows
			cols = scene.cols
			y = 0
			while y < rows:
				x = 0
				while x < cols:
					if scene.tiles[x][y].isBalloon:
						self.balloonCoords[(x, y)] = scene.tiles[x][y]
					x += 1
				y += 1
		
	def update(self):		
		if not self.done:
			player = self.scene.player
			tx = player.x // 16
			ty = player.y // 16
			
			if self.id == 'water1' and ty > 18: return
			if self.id == 'water4' and ty < 18: return
			
			tile = self.balloonCoords.get((tx, ty), None)
			if tile != None:
				self.applyOverlayAndSaveContext(True, (tx + tile.primaryBalloonOffset[0], ty + tile.primaryBalloonOffset[1]))

	def postInit(self):
		if self.id == 'water1':
			if self.context.balloonA:
				self.applyOverlayAndSaveContext()
		elif self.id == 'water2':
			if self.context.balloonB:
				self.applyOverlayAndSaveContext()
		elif self.id == 'water3':
			if self.context.balloonC:
				self.applyOverlayAndSaveContext()
		elif self.id == 'water4':
			if self.context.balloonD:
				self.applyOverlayAndSaveContext()
		elif self.id == 'lava1':
			if self.context.volcanoA:
				self.applyOverlayAndSaveContext()
		elif self.id == 'lava2':
			if self.context.volcanoB:
				self.applyOverlayAndSaveContext()
		elif self.id == 'lava3':
			if self.context.volcanoC:
				self.applyOverlayAndSaveContext()
	
	def applyOverlayAndSaveContext(self, showPop=False, balloonLoc=None):
		if self.done:
			return
		overlayMapName = self.scene.overlayTriggers[self.id] # Go ahead and crash if not present. Game is busted at this point anyway.
		
		overlay = PlayScene(overlayMapName, 0, 0, Context())
		original = self.scene
		
		width = min(original.cols, original.cols)
		height = min(original.rows, original.rows)
		grid = original.tiles
		overlayTiles = overlay.tiles
		y = 0
		while y < height:
			x = 0
			while x < width:
				if len(overlay.tiles[x][y].originalTemplates) > 0:
					templates = grid[x][y].originalTemplates + overlay.tiles[x][y].originalTemplates
					grid[x][y] = Tile(templates, x, y)
				x += 1
			y += 1
		
		if self.id == 'water1':
			self.context.balloonA = True
		elif self.id == 'water2':
			self.context.balloonB = True
		elif self.id == 'water3':
			self.context.balloonC = True
		elif self.id == 'water4':
			self.context.balloonD = True
		
		self.done = True
		
		if showPop:
			playNoise('water_pop')
			original.sprites.append(Sprite('waterpop', balloonLoc[0] * 16 + 8, balloonLoc[1] * 16 + 8))
		
		
			if self.context.balloonA and self.context.balloonB and self.context.balloonC and self.context.balloonD:
				self.scene.next = DialogScene(self.scene, 'WaterDone', True, False, False)

class BikeLevelStuff(SpecialLevelStuff):
	def __init__(self, scene):
		SpecialLevelStuff.__init__(self, scene)
		self.hasUpdate = True
		self.hasRender = True
		self.particles = []
		self.counter = 0
		
	def update(self):
		if self.counter == 0:
			self.scene.sprites.append(Sprite('bird2', 16, 48))
		self.counter += 1
		scene = self.scene
		player = scene.player
		if player.x > (scene.cols - 1) * 16: # one less because you can't technically go off screen
			scene.next = VictorySequenceScene(scene)
		
		if self.counter % 2 == 0:
			pt = StemCell()
			pt.x = player.x
			pt.y = player.y
			pt.vx = random.random() * 16 - 8
			pt.vy = random.random() * 6 - 5
			self.particles.append(pt)
		
		for pt in self.particles:
			pt.x += pt.vx
			pt.y += pt.vy
			pt.vy += 2
	
	def render(self, screen, rc, offsetX, offsetY):
		p = self.scene.player
		px = p.x + offsetX
		py = p.y + offsetY
		
		for pt in self.particles:
			PDR(screen, (255, 255, 128), pygame.Rect(pt.x + offsetX, pt.y + offsetY, 3, 3))
		
		if len(self.particles) > 40:
			self.particles = self.particles[1:]
		
		
		#screen.blit(getText(WHITE, "Don't worry, I'll redo the"), (0, 8))
		#screen.blit(getText(WHITE, "scatter pattern before"), (0, 16))
		#screen.blit(getText(WHITE, "we turn it in."), (0, 24))
		#screen.blit(getText(WHITE, "              --Blake"), (0, 32))

class BirdFeeder(SpecialLevelStuff):
	def __init__(self, scene):
		SpecialLevelStuff.__init__(self, scene)
		self.hasPostInit = True
		self.hasUpdate = True
		self.counter = 0
		seeds = [(4,36),
			(11,30),
			(4,24),
			(11,18),
			(4,12),
			(11,6)]
		self.seedsLeft = len(seeds)
		for seed in seeds:
			scene.sprites.append(Sprite('seeds', seed[0] * 16 + 8, seed[1] * 16 + 7))
		
		bike = Sprite('rocketbike', 12 * 16 + 8, 39 * 16 + 8)
		bike.floats = True
		scene.sprites.append(bike)
		
		self.closingShown = False
		
	def update(self):
		p = self.scene.player
		self.counter += 1
		
		if self.counter == 3:
			self.scene.next = DialogScene(self.scene, "Feeder1", False, False, False)
		
		if self.seedsLeft == 0 and not self.closingShown:
			self.scene.next = DialogScene(self.scene, "Feeder2", False, False, False)
			self.scene.sprites.append(Sprite('bird1', 8, 8))
			self.closingShown = True
		
		
		
	def postInit(self):
		pass
		


class GravityAscentExit(SpecialLevelStuff):
	def __init__(self, scene):
		SpecialLevelStuff.__init__(self, scene)
		self.hasDoorTrigger = True
	
	def doorTrigger(self, doorId):
		if doorId == 'XXXXX':
			self.scene.next = PlayScene('ship_1', 8, 9, self.scene.context)
			return None
		return doorId


class GravityCorePlacement(SpecialLevelStuff):
	def __init__(self, scene):
		SpecialLevelStuff.__init__(self, scene)
		self.hasPostInit = True
		self.hasUpdate = True
		self.col = 11
		self.row = 9
		self.done = False
	
	def update(self):
		player = self.scene.player
		tx = int(player.x / 16)
		ty = int(player.y / 16)
	
		if tx == self.col and abs(ty - self.row) < 2 and not self.done:
			self.addGDevice(True)
	
	def postInit(self):
		if self.context.gravity:
			self.addGDevice()
	
	def addGDevice(self, showDialog=False):
		self.done = True
		self.context.gravity = True
		gd = Sprite('gravity_device', self.col * 16, self.row * 16)
		gd.floats = True
		self.scene.sprites.append(gd)
		if showDialog:
			self.scene.next = DialogScene(self.scene, 'GravityDone', True, False, False)


class ShipTransmission(SpecialLevelStuff):
	def __init__(self, scene):
		SpecialLevelStuff.__init__(self, scene)
		self.hasUpdate = True
	
	def update(self):
		
		ctx = self.context
		if not ctx.gravity:
			if not ctx.transmission1:
				self.scene.triggerDialog('T1', True)
				ctx.transmission1 = True
		elif not (ctx.volcanoA and ctx.volcanoB and ctx.volcanoC):
			if not ctx.transmission2:
				self.scene.triggerDialog('T2', True)
				ctx.transmission2 = True
		elif not (ctx.balloonA and ctx.balloonB and ctx.balloonC):
			if not ctx.transmission3:
				self.scene.triggerDialog('T3', True)
				ctx.transmission3 = True
		else:
			if not ctx.transmission4:
				self.scene.triggerDialog('T4', True)
				ctx.transmission4 = True
	

class VolcanoCoreActivation(SpecialLevelStuff):
	def __init__(self, scene, id):
		SpecialLevelStuff.__init__(self, scene)
		self.id = id
		self.scene = scene
		self.context = scene.context
		self.hasUpdate = True
		self.hasPostInit = True
		self.spotX = 11
		self.spotY = 9
		self.sequenceCounter = -1
		self.done = False
		self.lavaLevel = 0
		self.shakeScreen = False
		self.freeze = False
		self.warpBack = False
		
	def update(self):
		
		player = self.scene.player
		tx = player.x // 16
		ty = player.y // 16
		
		if tx == self.spotX and abs(ty - self.spotY) < 2 and not self.done:
			self.doLavaSequence(False)
		
		self.freeze = False
		self.shakeScreen = False
		self.scene.player.sprinkle = False
		
		if self.sequenceCounter >= 0:
			SPRINKLE = 30 # player sprinkles mix
			SHAKE = 60 # screenshakes 
			RISE_RATE = 30 # this happens 3 times
			
			self.freeze = True
			sc = self.sequenceCounter
			if sc < SPRINKLE:
				self.scene.player.sprinkle = True
				if sc == 0:
					playNoise('sprinkle_lava_packet')
				progress = 1.0 * sc / SPRINKLE
			else:
				sc -= SPRINKLE
				if sc < SHAKE:
					if sc == 0:
						playNoise('screen_shaking')
					self.shakeScreen = (sc & 1) == 0
					
				else:
					sc -= SHAKE
					if sc < RISE_RATE:
						if sc == 0:
							playNoise('lava_rise')
						self.lavaLevel = 1
					else:
						sc -= RISE_RATE
						if sc < RISE_RATE:
							if sc == 0:
								playNoise('lava_rise')
							self.lavaLevel = 2
						else:
							sc -= RISE_RATE
							self.lavaLevel = 3
							if sc < RISE_RATE:
								if sc == 0:
									playNoise('lava_rise')
							else:
								self.freeze = False
								if self.warpBack:
									self.scene.next = VolcanoCompleteScene(self.scene, self.context)
			self.sequenceCounter += 1

	def postInit(self):
		isLavaOn = False
		if self.id == 1:
			isLavaOn = self.context.volcanoA
		elif self.id == 2:
			isLavaOn = self.context.volcanoB
		elif self.id == 3:
			isLavaOn = self.context.volcanoC
		
		if isLavaOn:
			self.doLavaSequence(True)
		
	
	def doLavaSequence(self, instant):
		self.done = True
		if instant:
			self.sequenceCounter = 9999999
		else:
			self.sequenceCounter = 0
			
		
		complete = self.context.volcanoA and self.context.volcanoB and self.context.volcanoC
		
		if not complete:
			if self.id == 1:
				self.context.volcanoA = True
			elif self.id == 2:
				self.context.volcanoB = True
			elif self.id == 3:
				self.context.volcanoC = True
			
			complete = self.context.volcanoA and self.context.volcanoB and self.context.volcanoC
			if complete:
				self.warpBack = True
			

NEIGHBOR_RANGE = (-1, 0, 1)

JUMPING_VY = -14
RUN_JUMPING_VY = -16
WATER_JUMPING_VY = -5

SPRITE_HEIGHT = {
	'player_side': 32,
	'acorn': 16,
	'acorntop': 16,
	'gravity_device': 16,
	'lazor': 16,
	'lavamonster': 32,
	'waterpop': 32,
	'powerup': 16,
	'poof_big': 32,
	'poof_small': 16,
	'byat': 16,
	'moonsquid': 16,
	'seed': 16,
	'rocketbike': 32,
}

G = 0.7
STRONG_G = 1.55
WATER_G = 0.1

def getSpriteHeight(type):
	return SPRITE_HEIGHT.get(type, 16)

XY_PAIRINGS = [
	(0, 0),
	(1, 0),
	(2, 0),
	(3, 0),
	(4, 0),
	(4, 1),
	(4, 2),
	(4, 3)
]

ENEMY_TYPES = [
	'acorn',
	'acorntop'
]

class Sprite:
	def __init__(self, type, px, py):
		self.type = type
		self.isEnemy = False
		self.height = getSpriteHeight(type)
		self.x = px
		self.y = py
		self.modelX = px + 0.0
		self.modelY = py + 0.0
		self.xs = [None] * 5
		self.ys = [None] * 5
		self.vy = 0
		
		self.moving = False
		self.lastDirection = 'left'
		self.onGround = False
		self.neighbors = [None] * 36
		self.spawns = None
		self.dead = False
		self.collidedWall = False
		self.deleteWhenOffScreen = False
		self.ghost = False
		self.renderImpl = SPRITE_renderPlayerOver
		self.automation = None
		self.floats = False
		self.deathState = None
		self.sprinkle = False
		self.isEnemy = type in ENEMY_TYPES
		self.isBullet = type == 'lazor'
		self.inBackground = False
		self.bikemode = False
		self.powerupInfo = None
		if type == 'acorn':
			self.renderImpl = SPRITE_renderAcorn
			self.automation = AcornAutomation(self)
		elif type == 'acorntop':
			self.renderImpl = SPRITE_renderAcornTop
			self.automation = AcornTopAutomation(self)
			self.acorntopdir = 'left'
		elif type == 'gravity_device':
			self.renderImpl = SPRITE_renderGravityDevice
			self.automation = None
		elif type == 'lazor':
			self.renderImpl = SPRITE_renderLazor
			self.automation = LazorAutomation(self)
		elif type == 'lavamonster':
			self.renderImpl = SPRITE_renderLavaMonster
			self.automation = LavaMonsterAutomation(self)
			self.inBackground = True
			self.floats = True
			self.ghost = True
		elif type == 'fireball':
			self.renderImpl = SPRITE_renderFireball
			self.automation = FireBallAutomation(self)
			self.floats = True
			# not a ghost, though. Collides with walls and that triggers it to fizzle.
		elif type == 'waterpop':
			self.renderImpl = SPRITE_renderWaterPop
			self.automation = WaterPopAutomation(self)
			self.floats = True
		elif type == 'powerup':
			self.renderImpl = SPRITE_renderPowerup
			self.automation = PowerupAutomation(self)
		elif type == 'poof_big':
			self.renderImpl = SPRITE_renderPoof
			self.automation = PoofAutomation(self, True)
			self.floats = True
		elif type == 'poof_small':
			self.renderImpl = SPRITE_renderPoof
			self.automation = PoofAutomation(self, False)
			self.floats = True
		elif type == 'byat':
			self.renderImpl = SPRITE_renderByat
			self.automation = ByatAutomation(self)
			self.floats = True
		elif type == 'moonsquid':
			self.renderImpl = SPRITE_renderMoonSquid
			self.automation = MoonSquidAutomation(self)
			self.floats = True
		elif type == 'seeds':
			self.renderImpl = SPRITE_renderSeed
			self.automation = SeedAutomation(self)
		elif type == 'bird1':
			self.renderImpl = SPRITE_renderBird
			self.automation = BirdAutomation(self, False)
			self.floats = True
		elif type == 'bird2':
			self.renderImpl = SPRITE_renderBird
			self.automation = BirdAutomation(self, True)
			self.floats = True
			self.ghost = True
		elif type == 'rocketbike':
			self.renderImpl = SPRITE_renderRocketBike
			self.automation = RocketBikeAutomation(self)
			self.ghost = True
			
		
		self.dx = 0
		self.dy = 0
		self.iceVX = 0
		self.ddx = 0 # "damage dx", will stay set until you land on the ground or blink counter goes < 0
		self.damageDir = None
		self.cling = False
		self.ladderDY = 0
		self.blinkCounter = -1
		self.cameFromIce = False
		
	def checkNeighborCollision(self, scene, col, row, targetX, targetY):
		area_left = targetX - 5
		area_right = targetX + 5
		area_top = targetY - 5
		area_bottom = targetY + 5
		
		width = scene.cols
		height = scene.rows
		left = col - 2
		right = col + 2
		top = row - 2
		bottom = row + 2
		if left < 0: left = 0
		if right >= width: right = width - 1
		if top < 0: top = 0
		if bottom >= height: bottom = height - 1
		index = 0
		y = top
		while y <= bottom:
			x = left
			while x <= right:
				cs = scene.tiles[x][y].collisions
				csi = len(cs) - 1
				while csi >= 0:
					self.neighbors[index] = cs[csi]
					csi -= 1
					index += 1
				x += 1
			y += 1
		
		index -= 1
		while index >= 0:
			n = self.neighbors[index]
			if n[0] > area_right:
				pass
			elif n[1] > area_bottom:
				pass
			elif n[2] < area_left:
				pass
			elif n[3] < area_top:
				pass
			else:
				return False
			index -= 1
		return True
		
	def update(self, scene):
		
		if self.automation != None:
			self.automation.doStuff(scene)
		
		self.blinkCounter -= 1
		if self.blinkCounter < 0 or self.onGround:
			self.ddx = 0
			if self.damageDir != None:
				self.lastDirection = self.damageDir
			self.damageDir = None
		
		if self.damageDir != None:
			self.dx = 0
		
		self.dx += self.ddx
		if scene.side:
			# hotspot is located in the center of the bottom most tile
			areaX = self.modelX
			areaBottom = self.modelY + 8
			heightFromHotSpot = getSpriteHeight(self.type) - 8 - 4
			areaTop = self.modelY - heightFromHotSpot
			waterOnly = self.type == 'moonsquid'
			
			width = scene.cols
			height = scene.rows
			
			isPlayer = self == scene.player 
			bikePhysics = isPlayer and scene.id == 'bike_level'
			
			if bikePhysics:
				offsetX = -(scene.updateCounter * BIKE_SPEED)
				if offsetX > 0: offsetX = 0
				right = -(scene.cols * 16 - 256)
				if offsetX < right: offsetX = right
				
				
				newX = areaX + self.dx
				pixelX = newX + offsetX
				if pixelX < 8:
					self.dx = BIKE_SPEED
				elif pixelX > 248:
					self.dx = BIKE_SPEED
				
			
			tileBottomIceCheck = int(areaBottom // 16)
			tileX = int(areaX // 16)
			inWater = False
			onIce = False
			if tileX >= 0 and tileBottomIceCheck > 0 and tileX < scene.cols and tileBottomIceCheck < scene.rows:
				tileBelow = scene.tiles[tileX][tileBottomIceCheck]
				if tileBelow != None:
					onIce = tileBelow.isIce
				
				inWater = scene.tiles[tileX][tileBottomIceCheck - 1].isWater
			
			ICE_DIMINISH = .1
			ICE_MAX_SPEED = 8
			
			if self.type != 'lazor' and (onIce or (self.cameFromIce and not self.onGround and not self.cling)):
				
				if self.dx == 0:
					if self.iceVX > 0:
						self.iceVX -= ICE_DIMINISH
						if self.iceVX < 0:
							self.iceVX = 0
					elif self.iceVX < 0:
						self.iceVX += ICE_DIMINISH
						if self.iceVX > 0:
							self.iceVX = 0
				elif self.dx > 0:
					self.iceVX += ICE_DIMINISH
				elif self.dx < 0:
					self.iceVX -= ICE_DIMINISH
				
				self.dx = self.iceVX
				if self.dx > ICE_MAX_SPEED:
					self.dx = ICE_MAX_SPEED
				elif self.dx < -ICE_MAX_SPEED:
					self.dx = -ICE_MAX_SPEED
				self.cameFromIce = True
			else:
				self.iceVX = 0
				if self.onGround:
					self.cameFromIce = False
			
			self.moving = self.dx != 0 or self.dy != 0
			self.collidedWall = False
			# side-to-side calcuation is done first, independent of whether you are on the ground.
			if self.dx != 0:
				if inWater and isPlayer:
					maxDX = 1.0 if self.onGround else 1.5
					if self.dx < -maxDX:
						self.dx = -maxDX
					elif self.dx > maxDX:
						self.dx = maxDX
					
					
				newX = self.modelX + self.dx
				
				# isCollision ignores collisions near the sprite's feet
				# if you are near ground, you will be automatically placed standing
				# upon it at the end of the vertical adjustment phase.
				if self.ghost or not scene.isCollision(newX, areaTop, newX, areaBottom):
				
					if self.cling:
						if int(newX // 16) != int(self.modelX // 16):
							newX = self.modelX
					self.modelX = newX
					if self.dx > 0:
						self.lastDirection = 'right'
					else:
						self.lastDirection = 'left'
				else:
					self.iceVX = 0
					self.collidedWall = not self.ghost
				self.dx = 0
			
			tileX = int(self.modelX / 16)
			tileY = int(self.modelY / 16)
			
			offScreen = tileX < 0 or tileX >= scene.cols or tileY < 0 or tileY >= scene.rows
			onScreen = not offScreen
			
			if bikePhysics and self.collidedWall:
				scene.next = DeathOverrideScene(scene, 'lava')
			
			# vertical adjustment phase
			# assume sprite is flying through the air unless you see ground
			wasOnGround = self.onGround # save the previous ground state. If you weren't on ground before but suddenly are, then you "landed" and a sound should be played. 
			self.onGround = False
			
			if wasOnGround or self.floats:
				self.vy = 0
			else:
				if inWater:
					self.vy += WATER_G
				elif scene.context.gravity:
					self.vy += STRONG_G
				else:
					self.vy += G
			
			rocketJump = False
			if inWater and isPlayer and self.vy < 0:
				tileY = int(self.modelY / 16) - 1
				if tileY >= 0:
					if not scene.tiles[tileX][tileY].isWater:
						# moving up, bottom half is in water, top half is not
						rocketJump = True # jump out with high velocity
			
			# If you're clinging to a ladder, then throw out the vy entirely.
			# Use self.ladderDY instead
			if self.cling:
				self.dy = self.ladderDY
				self.vy = 0
				if self.dy != 0:
					self.moving = True
			else:
				self.dy += self.vy
			
			oldLadderDY = self.ladderDY
			self.ladderDY = 0
			
			maxDY = 10
			if self.dy > 4 and inWater:
				self.dy = 4
			
			if self.dy > maxDY:
				self.dy = maxDY
			elif self.dy < -maxDY:
				self.dy = -maxDY
			
			
			if rocketJump:
				self.dy = -15
			
			
			wasCling = self.cling
			movedUp = False
			
			newBottom = areaBottom + self.dy
			newTop = areaTop + self.dy
			
			no = False
			
			if self.ghost:
				pass # just go with it
			else:
				if self.dy < 0:
					# Going Up
					if newTop < 0:
						# TODO: do top transitions here.
						no = True
					else:
						newTileTop = int(newTop / 16)
						
						offScreen = tileX < 0 or tileX >= scene.cols or newTileTop < 0 or newTileTop >= scene.rows
						onScreen = not offScreen
						
						if onScreen:
							t = scene.tiles[tileX][newTileTop]
							if t.solid or (waterOnly and not t.isWater):
								if wasOnGround:
									self.onGround = True
								no = True
								
								if self.modelY % 16 == 8:
									no = False
								self.vy = 0
								self.dy = 0
								if self == scene.player:
									playNoise('head_bonk')
							else:
								movedUp = True
						else:
							movedUp = True
				else:
					# Going Down (or staying the same)
					# The philosophy here is different. When you encounter a collision,
					# rather than stop the movement, you find where the top of the collided
					# tile is and place the sprite there.
					
					newTileBottom = int(newBottom / 16)
					
					# If the player fell off the bottom
					if isPlayer and newTileBottom >= height:
						# TODO: check if there is a transition registered with the map.
						# If so begin that transition instead of killing the player.
						playNoise('fall_to_death')
						scene.next = DeathOverrideScene(scene, 'fall')
						no = True
					else:
					
						offScreen = tileX < 0 or tileX >= scene.cols or newTileBottom < 0 or newTileBottom >= scene.rows
						onScreen = not offScreen
						
						if onScreen:
							t = scene.tiles[tileX][newTileBottom]
							topStop = False
							if t.isTop:
								oldTileBottom = int(areaBottom / 16)
								if newTileBottom > oldTileBottom:
									topStop = True
								elif newTileBottom == oldTileBottom and areaBottom == newBottom:
									topStop = True
							if t.solid or topStop:
								no = True
								self.cling = False
								self.onGround = True
								self.modelY = newTileBottom * 16 - 8
								self.vy = 0
								if not wasOnGround and self == scene.player:
									playNoise('land_on_ground')
						else:
							no = True
					
			if not no:
				self.modelY = newBottom - 8
			self.x = int(self.modelX)
			self.y = int(self.modelY)
			self.dy = 0
			self.dy = 0
			
			tileY = int(self.modelY / 16)
			
			offScreen = tileX < 0 or tileX >= scene.cols or tileY < 0 or tileY >= scene.rows
			onScreen = not offScreen
			if onScreen and scene.tiles[tileX][tileY].isLadder:
				pass
			else:
				self.cling = False
			
			if movedUp and wasCling and not self.cling:
				self.modelY = 16 * tileY # maybe?
			
		else:
			self.moving = False
			if self.dx != 0 or self.dy != 0:
				self.moving = True
				
				if self.dy < 0:
					self.lastDirection = 'up'
				elif self.dy > 0:
					self.lastDirection = 'down'
				elif self.dx < 0:
					self.lastDirection = 'left'
				else:
					self.lastDirection = 'right'
					
				xs = self.xs
				ys = self.ys
				
				xs[0] = self.modelX + self.dx
				xs[1] = self.modelX + self.dx * 3.0 / 4
				xs[2] = self.modelX + self.dx / 2.0
				xs[3] = self.modelX + self.dx / 4.0
				xs[4] = self.modelX
				
				ys[0] = self.modelY + self.dy
				ys[1] = self.modelY + self.dy * 3.0 / 4
				ys[2] = self.modelY + self.dy / 2.0
				ys[3] = self.modelY + self.dy / 4.0

				for pairing in XY_PAIRINGS:
					newx = xs[pairing[0]]
					newy = ys[pairing[1]]
					col = int(newx / 16)
					row = int(newy / 16)

					if self.checkNeighborCollision(scene, col, row, newx, newy):
						self.modelX = newx
						self.modelY = newy
						self.x = int(self.modelX)
						self.y = int(self.modelY)
						break
				
				self.dx = 0
				self.dy = 0
	
	def render(self, scene, screen, offsetX, offsetY, rc):
		if self.blinkCounter < 0 or (self.blinkCounter & 1) == 0:
			self.renderImpl(self, scene, screen, offsetX, offsetY, rc)
	
	def isCollision(self, other):
		left = self.x - 8
		oleft = other.x - 8
		right = left + 16
		oright = oleft + 16
		
		if left >= oright or right <= oleft:
			return False
		
		bottom = self.y + 8
		obottom = other.y + 8
		top = bottom - self.height
		otop = obottom - other.height
		
		if top >= obottom or otop >= bottom:
			return False
		return True
	
	def hit(self, scene, amount):
		self.blinkCounter = 40
		scene.context.adjustHealth(scene, -amount)
		self.blinkDirection = self.lastDirection
		
		if scene.id != 'bike_level':
		
			tiles = scene.tiles
			tx = self.x // 16
			ty = self.y // 16
			
			inWater = False
			if tx >= 0 and ty >= 0 and tx < scene.cols and ty < scene.rows:
				t = tiles[tx][ty]
				inWater = t.isWater
			
			self.damageDir = self.lastDirection
			if self.lastDirection == 'left':
				self.ddx = 3
			else:
				self.ddx = -3
			self.onGround = False
			if inWater:
				self.vy = -1
			else:
				self.vy = -6


def SPRITE_renderAcorn(sprite, scene, screen, offsetX, offsetY, arc):
	left = sprite.x + offsetX - 8
	top = sprite.y + offsetY - 8
	width = 16
	height = 16

	base = 'acorn'
	if sprite.automation.mode in ('throwing', 'baldpatrolling'):
		base = 'acorn_capless'
	
	if sprite.moving:
		path = base + '_' + '13'[(arc // 2) & 1]
	else:
		path = base + '_2'
	
	path = 'sprites/' + path + '.png'
	reverse = sprite.lastDirection == 'left'
	if reverse:
		img = getBackwardsImage(path)
	else:
		img = getImage(path)
	screen.blit(img, (left, top))



def SPRITE_renderAcornTop(sprite, scene, screen, offsetX, offsetY, arc):
	left = sprite.x + offsetX - 8
	top = sprite.y + offsetY - 8
	width = 16
	height = 16
	path = 'acorn_projectile_1'
	path = 'sprites/' + path + '.png'
	img = getImage(path)
	screen.blit(img, (left, top))



def SPRITE_renderBird(sprite, scene, screen, offsetX, offsetY, arc):
	left = sprite.x + offsetX - 16
	top = sprite.y + offsetY - 16
	img = getImage('sprites/bird_chase_' + ('1' if (((arc // 2) & 1) == 0) else '2') + '.png')
	screen.blit(img, (left, top))



def SPRITE_renderByat(sprite, scene, screen, offsetX, offsetY, arc):
	left = sprite.x + offsetX - 8
	top = sprite.y + offsetY - 8
	if sprite.automation.perched:
		path = 'sprites/byat_hang.png'
	else:
		if (arc & 1) == 0:
			path = 'sprites/byat_fly_1.png'
		else:
			path = 'sprites/byat_fly_2.png'
	
	if sprite.lastDirection == 'left':
		img = getBackwardsImage(path)
	else:
		img = getImage(path)
		
	screen.blit(img, (left, top))


def SPRITE_renderFireball(sprite, scene, screen, offsetX, offsetY, arc):

	left = sprite.x - 8 + offsetX
	top = sprite.y - 8 + offsetY
	
	img = getImage('sprites/lava_monster_projectile.png')
	
	screen.blit(img, (left, top))


def SPRITE_renderGravityDevice(sprite, scene, screen, offsetX, offsetY, arc):
	left = sprite.x + offsetX - 8
	top = sprite.y + offsetY - 8
	i = str((arc % 3) + 1)
	screen.blit(getImage('sprites/gravity_device_' + i + '.png'), (left, top))


def SPRITE_renderLavaMonster(sprite, scene, screen, offsetX, offsetY, arc):

	left = sprite.x - 8 + offsetX
	top = sprite.y - 24 + offsetY
	
	
	if sprite.automation.mouthOpen:
		path = 'sprites/lava_monster_2.png'
	else:
		path = 'sprites/lava_monster_1.png'
	if sprite.lastDirection == 'left':
		img = getBackwardsImage(path)
	else:
		img = getImage(path)
	
	screen.blit(img, (left, top))

PDR = pygame.draw.rect
YELLOW = (255, 255, 0)
def SPRITE_renderLazor(sprite, scene, screen, offsetX, offsetY, arc):
	x = offsetX + sprite.x - 8
	y = offsetY + sprite.y - 2
	PDR(screen, YELLOW, pygame.Rect(x, y, 16, 4))

def SPRITE_renderMoonSquid(sprite, scene, screen, offsetX, offsetY, arc):
	left = offsetX + sprite.x - 8
	top = offsetY + sprite.y - 8
	arc = arc // 3
	img = getImage('sprites/moonsquid_1.png' if ((arc & 1) == 0) else 'sprites/moonsquid_2.png')
	screen.blit(img, (left, top))
	


#arc = adjusted render counter (slowed down for animation frames, so I don't have to do rc = (rc // 4) to slow things down
def SPRITE_renderPlayerOver(sprite, scene, screen, offsetX, offsetY, arc):
	left = sprite.x + offsetX - 8
	top = sprite.y + offsetY - 8
	width = 16
	height = 16
	gun = scene.lazor_cooldown > 0
	if scene.side:
		top -= 16
		height = 32
	if sprite.bikemode:
		if (arc & 1) == 0:
			img = getImage('sprites/rocket_bike_1.png')
		else:
			img = getImage('sprites/rocket_bike_2.png')
	elif scene.side:
	
		if sprite.damageDir != None:
			dir = sprite.damageDir
		else:
			dir = sprite.lastDirection
		reverse = dir == 'left'
		
		base = 'cave' if scene.hasAtmosphere else 'space'
		moving = sprite.moving
		if sprite.cling:
			path = base + '_climb_'
			if gun:
				path += 'shoot'
				if reverse:
					left -= 8
				else:
					left += 8
			elif moving:
				path += '1232'[arc % 4]
			else:
				path += '2'
		else:
			path = base + '_'
			if gun:
				if moving:
					path += 'shooting_' + '1213'[arc % 4]
				else:
					path += 'shooting'
			else:
				if moving or (sprite.cling and sprite.vy != 0):
					# TODO: running, cut RC in half
					path += '1213'[arc % 4]
				else:
					path += '1'
		
		x = left
		y = top
		if sprite.deathState != None and sprite.deathState != 'fall':
			if sprite.deathState == 'collapse':
				img = getImage('sprites/' + base + '_death.png')
				top += 16
				left -= 8
			elif sprite.deathState == 'lava':
				if reverse:
					img = getBackwardsImage('sprites/roasted.png')
				else:
					img = getImage('sprites/roasted.png')
					left = left + 16 - img.get_width()
		else:
			if sprite.sprinkle:
				img = getImage('sprites/space_sprinkle.png')
				left = left + 16 - img.get_width()
			elif reverse:
				img = getBackwardsImage('sprites/' + path +'.png')
			else:
				img = getImage('sprites/' + path +'.png')
				left = left + 16 - img.get_width()
			
	else:
		counter = '1232'[arc % 4] if sprite.moving else '2'
		dir = sprite.lastDirection
		reverse = False
		if dir == 'left':
			reverse = True
			dir = 'right'
		base = 'space'
		ctx = scene.context
		if ctx.volcanoA and ctx.volcanoB and ctx.volcanoC:
			base = 'cave'
		path = 'sprites/' + base + '_overworld_' + dir + '_' + counter + '.png'
		img = getBackwardsImage(path) if reverse else getImage(path) 
	screen.blit(img, (left, top))



def SPRITE_renderPoof(sprite, scene, screen, offsetX, offsetY, arc):
	isBig = sprite.automation.isBig
	left = sprite.x + offsetX - 8
	top = sprite.y + offsetY - 8
	
	c = sprite.automation.counter
	if c <= 3:
		num = '1'
	elif c <= 6:
		num = '2'
	elif c <= 9:
		num = '3'
	else:
		return
	
	if isBig:
		top -= 16
		img = getImage('sprites/large_poof_' + num + '.png')
	else:
		img = getImage('sprites/small_poof_' + num + '.png')
	
	screen.blit(img, (left, top))



def SPRITE_renderPowerup(sprite, scene, screen, offsetX, offsetY, arc):
	
	if sprite.automation.visible:
		
		left = sprite.x + offsetX - 8
		top = sprite.y + offsetY - 8
		type = sprite.powerupInfo.type
		if type == '1up':
			path = 'sprites/1up.png'
		elif type == 'life_small':
			path = 'sprites/life_small.png'
		elif type == 'life_big':
			path = 'sprites/life_big.png'
		
		screen.blit(getImage(path), (left, top))



def SPRITE_renderRocketBike(sprite, scene, screen, offsetX, offsetY, arc):
	left = sprite.x + offsetX - 16
	top = sprite.y + offsetY - 24
	img = getImage('sprites/rocket_bike_squirrelless.png')
	screen.blit(img, (left, top))



def SPRITE_renderSeed(sprite, scene, screen, offsetX, offsetY, arc):
	left = sprite.x + offsetX - 8
	top = sprite.y + offsetY - 8
	img = getImage('sprites/bird_seed.png')
	screen.blit(img, (left, top))



def SPRITE_renderWaterPop(sprite, scene, screen, offsetX, offsetY, arc):
	
	left = sprite.x + offsetX - 24
	top = sprite.y + offsetY - 8
	screen.blit(getImage('tiles/balloons/pop.png'), (left, top))


chars = {}
chars['^'] = [[0,1,1,0,0,0,0,0],[0,1,1,1,0,0,0,0],[0,1,1,1,1,0,0,0],[0,1,1,1,1,1,0,0],[0,1,1,1,1,1,0,0],[0,1,1,1,1,0,0,0],[0,1,1,1,0,0,0,0],[0,1,1,0,0,0,0,0]]
chars['!'] = [[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,1,0],[0,0,1,1,1,0,1,0],[1,1,1,1,0,0,0,0],[1,1,1,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['\''] = [[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[1,1,0,1,0,0,0,0],[1,1,1,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['"'] = [[0,0,0,0,0,0,0,0],[1,1,0,1,0,0,0,0],[1,1,1,0,0,0,0,0],[0,0,0,0,0,0,0,0],[1,1,0,1,0,0,0,0],[1,1,1,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['`'] = [[0,0,0,0,0,0,0,0],[0,1,1,1,0,0,0,0],[1,0,1,1,0,0,0,0],[0,0,0,0,0,0,0,0],[0,1,1,1,0,0,0,0],[1,0,1,1,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars[','] = [[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,1,1,0,1],[0,0,0,0,1,1,1,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['-'] = [[0,0,0,1,0,0,0,0],[0,0,0,1,0,0,0,0],[0,0,0,1,0,0,0,0],[0,0,0,1,0,0,0,0],[0,0,0,1,0,0,0,0],[0,0,0,1,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['.'] = [[0,0,0,0,0,0,0,0],[0,0,0,0,0,1,1,0],[0,0,0,0,0,1,1,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['/'] = [[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,1,0],[0,0,0,0,0,1,0,0],[0,0,0,0,1,0,0,0],[0,0,0,1,0,0,0,0],[0,0,1,0,0,0,0,0],[0,1,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['0'] = [[0,1,1,1,1,1,0,0],[1,1,1,1,1,1,1,0],[1,0,0,0,0,0,1,0],[1,0,0,0,0,0,1,0],[1,0,0,0,0,0,1,0],[1,1,1,1,1,1,1,0],[0,1,1,1,1,1,0,0],[0,0,0,0,0,0,0,0]]
chars['1'] = [[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,1,0,0,0,0,0,0],[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['2'] = [[0,1,0,0,0,1,1,0],[1,1,0,0,1,1,1,0],[1,0,0,1,1,0,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,1,1,1,0,0,1,0],[0,1,1,0,0,0,1,0],[0,0,0,0,0,0,0,0]]
chars['3'] = [[0,1,0,0,0,1,0,0],[1,1,0,0,0,1,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,1,1,1,1,1,1,0],[0,1,1,0,1,1,0,0],[0,0,0,0,0,0,0,0]]
chars['4'] = [[1,1,1,1,1,0,0,0],[1,1,1,1,1,0,0,0],[0,0,0,0,1,0,0,0],[0,0,0,0,1,0,0,0],[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[0,0,0,0,1,0,0,0],[0,0,0,0,0,0,0,0]]
chars['5'] = [[1,1,1,0,0,1,0,0],[1,1,1,0,0,1,1,0],[1,0,1,0,0,0,1,0],[1,0,1,0,0,0,1,0],[1,0,1,0,0,0,1,0],[1,0,1,1,1,1,1,0],[1,0,0,1,1,1,0,0],[0,0,0,0,0,0,0,0]]
chars['6'] = [[0,1,1,1,1,1,0,0],[1,1,1,1,1,1,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,1,0,1,1,1,1,0],[0,1,0,0,1,1,0,0],[0,0,0,0,0,0,0,0]]
chars['7'] = [[1,0,0,0,0,0,0,0],[1,0,0,0,0,0,0,0],[1,0,0,0,1,1,1,0],[1,0,0,1,1,1,1,0],[1,0,1,1,1,1,1,0],[1,1,1,0,0,0,0,0],[1,1,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['8'] = [[0,1,1,0,1,1,0,0],[1,1,1,1,1,1,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,1,1,1,1,1,1,0],[0,1,1,0,1,1,0,0],[0,0,0,0,0,0,0,0]]
chars['9'] = [[0,1,1,0,0,1,0,0],[1,1,1,1,0,1,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,1,1,1,1,1,1,0],[0,1,1,1,1,1,0,0],[0,0,0,0,0,0,0,0]]
chars[':'] = [[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,1,1,0,0,1,1,0],[0,1,1,0,0,1,1,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars[';'] = [[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[1,1,0,1,1,0,1,0],[1,1,0,1,1,1,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['<'] = [[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,1,1,0,0,0],[0,0,1,1,1,1,0,0],[0,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,1],[1,1,1,1,1,1,1,1],[0,0,0,0,0,0,0,0]]
chars['='] = [[0,0,0,0,0,0,0,0],[0,0,0,1,0,1,0,0],[0,0,0,1,0,1,0,0],[0,0,0,1,0,1,0,0],[0,0,0,1,0,1,0,0],[0,0,0,1,0,1,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['>'] = [[0,0,0,0,0,0,0,0],[1,1,1,1,1,1,1,1],[1,1,1,1,1,1,1,1],[0,1,1,1,1,1,1,0],[0,0,1,1,1,1,0,0],[0,0,0,1,1,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['?'] = [[0,0,0,0,0,0,0,0],[0,1,0,0,0,0,0,0],[1,1,0,0,0,0,1,0],[1,0,0,0,1,0,1,0],[1,0,0,1,1,0,0,0],[1,1,1,1,0,0,0,0],[0,1,1,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['*'] = [[0,0,1,1,1,1,0,0],[0,1,0,0,0,0,1,0],[1,0,0,1,1,0,0,1],[1,0,1,0,0,1,0,1],[1,0,1,0,0,1,0,1],[1,0,1,0,0,1,0,1],[0,1,0,0,0,0,1,0],[0,0,1,1,1,1,0,0]]
chars['A'] = [[0,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[1,0,0,0,1,0,0,0],[1,0,0,0,1,0,0,0],[1,0,0,0,1,0,0,0],[1,1,1,1,1,1,1,0],[0,1,1,1,1,1,1,0],[0,0,0,0,0,0,0,0]]
chars['B'] = [[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,1,1,1,1,1,1,0],[0,1,1,0,1,1,0,0],[0,0,0,0,0,0,0,0]]
chars['C'] = [[0,1,1,1,1,1,0,0],[1,1,1,1,1,1,1,0],[1,0,0,0,0,0,1,0],[1,0,0,0,0,0,1,0],[1,0,0,0,0,0,1,0],[1,1,0,0,0,1,1,0],[0,1,0,0,0,1,0,0],[0,0,0,0,0,0,0,0]]
chars['D'] = [[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[1,0,0,0,0,0,1,0],[1,0,0,0,0,0,1,0],[1,0,0,0,0,0,1,0],[1,1,1,1,1,1,1,0],[0,1,1,1,1,1,0,0],[0,0,0,0,0,0,0,0]]
chars['E'] = [[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,0,0,0,0,0,1,0],[0,0,0,0,0,0,0,0]]
chars['F'] = [[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[1,0,0,1,0,0,0,0],[1,0,0,1,0,0,0,0],[1,0,0,1,0,0,0,0],[1,0,0,1,0,0,0,0],[1,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['G'] = [[0,1,1,1,1,1,0,0],[1,1,1,1,1,1,1,0],[1,0,0,0,0,0,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,1,0,1,1,1,1,0],[0,1,0,1,1,1,0,0],[0,0,0,0,0,0,0,0]]
chars['H'] = [[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[0,0,0,1,0,0,0,0],[0,0,0,1,0,0,0,0],[0,0,0,1,0,0,0,0],[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[0,0,0,0,0,0,0,0]]
chars['I'] = [[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['J'] = [[0,0,0,0,1,1,0,0],[0,0,0,0,1,1,1,0],[0,0,0,0,0,0,1,0],[0,0,0,0,0,0,1,0],[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['K'] = [[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[0,0,0,1,0,0,0,0],[0,0,1,1,1,0,0,0],[0,1,1,0,1,1,0,0],[1,1,0,0,0,1,1,0],[1,0,0,0,0,0,1,0],[0,0,0,0,0,0,0,0]]
chars['L'] = [[0,0,0,0,0,0,0,0],[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[0,0,0,0,0,0,1,0],[0,0,0,0,0,0,1,0],[0,0,0,0,0,0,1,0],[0,0,0,0,0,0,1,0],[0,0,0,0,0,0,0,0]]
chars['M'] = [[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[0,1,1,0,0,0,0,0],[0,0,1,1,0,0,0,0],[0,1,1,0,0,0,0,0],[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[0,0,0,0,0,0,0,0]]
chars['N'] = [[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[0,1,1,0,0,0,0,0],[0,0,1,1,0,0,0,0],[0,0,0,1,1,0,0,0],[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[0,0,0,0,0,0,0,0]]
chars['O'] = [[0,1,1,1,1,1,0,0],[1,1,1,1,1,1,1,0],[1,0,0,0,0,0,1,0],[1,0,0,0,0,0,1,0],[1,0,0,0,0,0,1,0],[1,1,1,1,1,1,1,0],[0,1,1,1,1,1,0,0],[0,0,0,0,0,0,0,0]]
chars['P'] = [[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[1,0,0,0,1,0,0,0],[1,0,0,0,1,0,0,0],[1,0,0,0,1,0,0,0],[1,1,1,1,1,0,0,0],[0,1,1,1,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['Q'] = [[0,1,1,1,1,1,0,0],[1,1,1,1,1,1,1,0],[1,0,0,0,0,0,1,0],[1,0,0,0,1,0,1,0],[1,0,0,0,1,1,1,0],[1,1,1,1,1,1,0,0],[0,1,1,1,1,0,1,0],[0,0,0,0,0,0,0,0]]
chars['R'] = [[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[1,0,0,0,1,0,0,0],[1,0,0,0,1,0,0,0],[1,0,0,0,1,1,0,0],[1,1,1,1,1,1,1,0],[0,1,1,1,0,0,1,0],[0,0,0,0,0,0,0,0]]
chars['S'] = [[0,1,1,0,0,1,0,0],[1,1,1,1,0,1,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,0,0,1,0,0,1,0],[1,1,0,1,1,1,1,0],[0,1,0,0,1,1,0,0],[0,0,0,0,0,0,0,0]]
chars['T'] = [[1,0,0,0,0,0,0,0],[1,0,0,0,0,0,0,0],[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[1,0,0,0,0,0,0,0],[1,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['U'] = [[1,1,1,1,1,1,0,0],[1,1,1,1,1,1,1,0],[0,0,0,0,0,0,1,0],[0,0,0,0,0,0,1,0],[0,0,0,0,0,0,1,0],[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,0,0],[0,0,0,0,0,0,0,0]]
chars['V'] = [[1,1,1,1,0,0,0,0],[1,1,1,1,1,0,0,0],[0,0,0,0,1,1,0,0],[0,0,0,0,0,1,1,0],[0,0,0,0,1,1,0,0],[1,1,1,1,1,0,0,0],[1,1,1,1,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['W'] = [[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[0,0,0,0,1,1,0,0],[0,0,0,1,1,0,0,0],[0,0,0,0,1,1,0,0],[1,1,1,1,1,1,1,0],[1,1,1,1,1,1,1,0],[0,0,0,0,0,0,0,0]]
chars['X'] = [[1,1,0,0,0,1,1,0],[1,1,1,0,1,1,1,0],[0,0,1,1,1,0,0,0],[0,0,1,1,1,0,0,0],[1,1,1,0,1,1,1,0],[1,1,0,0,0,1,1,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['Y'] = [[1,1,1,0,0,0,0,0],[1,1,1,1,0,0,0,0],[0,0,0,1,1,1,1,0],[0,0,0,1,1,1,1,0],[1,1,1,1,0,0,0,0],[1,1,1,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['Z'] = [[1,1,0,0,0,1,1,0],[1,0,0,0,1,1,1,0],[1,0,0,1,1,0,1,0],[1,0,1,1,0,0,1,0],[1,1,1,0,0,0,1,0],[1,1,0,0,0,1,1,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['('] = [[1,1,1,1,1,0,0,0],[1,0,0,0,0,0,0,0],[1,0,0,0,0,0,0,0],[1,0,0,0,0,0,0,0],[1,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0]]
chars['&'] = [[0,0,0,1,1,1,1,0],[0,0,0,0,1,1,1,0],[0,0,0,1,1,0,0,0],[0,0,0,1,1,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,1,1,0],[0,0,0,0,0,1,1,0],[0,0,0,0,0,0,0,0]]
chars[')'] = [[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0],[0,0,0,0,0,0,1,0],[0,0,0,0,0,0,1,0],[0,0,0,0,0,0,1,0],[0,0,0,0,0,0,1,0],[0,0,1,1,1,1,1,0],[0,0,0,0,0,0,0,0]]

class Text:
	def __init__(self, chars):
		self.chars = chars
		self.colors = {}
		self.cache = {}
	
	def render(self, color, text):
		key = str(color) + '|' + text
		img = self.cache.get(key)
		if img == None:
			colorkey = str(color)
			colorlookup = self.colors.get(colorkey)
			if colorlookup == None:
				colorlookup = self.makeColorLookup(color)
				self.colors[colorkey] = colorlookup
			img = pygame.Surface((len(text) * 8, 8), pygame.SRCALPHA)
			
			x = 0
			for char in text:
				letter = colorlookup.get(char.upper())
				if letter != None:
					img.blit(letter, (x, 0))
				x += 8
			self.cache[key] = img

		return img
	
	def makeColorLookup(self, color):
		lookup = {}
		pixel = pygame.Surface((1, 1))
		pixel.fill(color)
		for key in self.chars.keys():
			grid = self.chars[key]
			img = pygame.Surface((8, 8), pygame.SRCALPHA)
			y = 0
			while y < 8:
				x = 0
				while x < 8:
					if grid[x][y] == 1:
						img.blit(pixel, (x, y))
					x += 1
				y += 1
			lookup[key] = img
		return lookup

_textEngine = Text(chars)

def renderText(color, text):
	return _textEngine.render(color, text)

# Because I keep accidentally using this one
getText = renderText


def clearTextCache(r, g, b):
	_textEngine.colors[str((r, g, b))] = {}


_tileStore = {}
def getTile(id):
	if len(_tileStore) == 0:
		initTileStore()
	
	return _tileStore.get(id)

def initTileStore():
	c = open('images/tiles/manifest.txt'.replace('/', os.sep), 'rt')
	t = c.read()
	c.close()
	
	for line in t.split('\n'):
		parts = trim(line.split('#')[0]).split('\t')
		if len(parts) == 3:
			id = trim(parts[0])
			flags = trim(parts[1])
			imagePaths = trim(parts[2]).split(',')
			
			tile = TileTemplate(id, flags, imagePaths)
			_tileStore[id] = tile
			
		
class TileTemplate:
	def __init__(self, id, flags, imagePaths):
		self.id = id
		self.solid = False
		self.coveredA = False
		self.coveredB = False
		self.coveredC = False
		self.coveredD = False
		self.isDoor = False
		self.isLadder = False
		self.isTop = False
		self.isWater = id in ('water', 'watertop')
		self.isSpike = False
		self.isLava = False
		self.isIce = False
		self.isBalloon = False
		self.primaryBalloonOffset = None # Location of top-middle balloon tile relative to this

		for flag in flags:
			if flag == 'x':
				self.solid = True
			elif flag == 'A':
				self.coveredA = True
			elif flag == 'B':
				self.coveredB = True
			elif flag == 'C':
				self.coveredC = True
			elif flag == 'D':
				self.coveredD = True
			elif flag == 'r' or flag == 'R':
				self.isDoor = True
			elif flag == 'L':
				self.isLadder = True
			elif flag == 'J':
				self.isTop = True
			elif flag == 'S':
				self.isSpike = True
			elif flag == 'V':
				self.isLava = True
			elif flag == 'I':
				self.isIce = True
			elif flag == 'o':
				self.isBalloon = True
				bid = int(id[-1])
				pOffset = (0, 0)
				if bid == 1:
					pOffset = (1, 0)
				elif bid == 3:
					pOffset = (-1, 0)
				elif bid == 4:
					pOffset= (1, -1)
				elif bid == 5:
					pOffset = (0, -1)
				elif bid == 6:
					pOffset = (-1, -1)
				self.primaryBalloonOffset = pOffset
		self.images = []
		for path in imagePaths:
			self.images.append(getImage('tiles/' + path))
		self.staticImage = None
		if len(self.images) == 1:
			self.staticImage = self.images[0]
		self.imageCount = len(self.images)
	
	def getImage(self, rc):
		img = self.staticImage
		if img == None:
			return self.images[(rc // 4) % self.imageCount]
		return img

def makeTile(lower, upper, col, row):
	if lower == None and upper == None:
		return Tile([], col, row)
	if lower == None:
		return Tile([upper], col, row)
	if upper == None:
		return Tile([lower], col, row)
	return Tile([lower, upper], col, row)

class Tile:
	def __init__(self, templates, col, row):
		self.col = col
		self.row = row
		self.originalTemplates = templates
		self.templates = []
		self.isDoor = False
		self.isLadder = False
		self.isTop = False
		self.isLava = False
		self.isWater = False
		self.isIce = False
		self.door = None
		self.isSpike = False
		self.isBalloon = False
		self.primaryBalloonOffset = None
		
		for tile in templates:
			if tile != None:
				self.templates.append(tile)
				if tile.isDoor:
					self.isDoor = True
				if tile.isLadder:
					self.isLadder = True
				if tile.isTop:
					self.isTop = True
				if tile.isWater:
					self.isWater = True
				if tile.isSpike:
					self.isSpike = True
				if tile.isLava:
					self.isLava = True
				if tile.isIce:
					self.isIce = True
				if tile.isBalloon:
					self.isBalloon = True
					self.primaryBalloonOffset = tile.primaryBalloonOffset
		
		self.collisions = []
		self.solid = False
		
		coveredA = False
		coveredB = False
		coveredC = False
		coveredD = False
		
		for t in self.templates:
			if t.solid:
				coveredA = True
				coveredB = True
				coveredC = True
				coveredD = True
				self.solid = True
			else:
				coveredA = coveredA or t.coveredA
				coveredB = coveredB or t.coveredB
				coveredC = coveredC or t.coveredC
				coveredD = coveredD or t.coveredD
		
		leftCovered = coveredA and coveredC
		rightCovered = coveredB and coveredD
		topCovered = coveredA and coveredB
		bottomCovered = coveredC and coveredD
		
		if leftCovered and rightCovered:
			self.collisions.append([0, 0, 2, 2])
			coveredA = False
			coveredB = False
			coveredC = False
			coveredD = False
			self.solid = True
		elif leftCovered:
			self.collisions.append([0, 0, 1, 2])
			coveredA = False
			coveredC = False
		elif rightCovered:
			self.collisions.append([1, 0, 1, 2])
			coveredB = False
			coveredD = False
		elif topCovered:
			self.collisions.append([0, 0, 2, 1])
			coveredA = False
			coveredB = False
		elif bottomCovered:
			self.collisions.append([0, 1, 2, 1])
			coveredC = False
			coveredD = False
		
		if coveredA:
			self.collisions.append([0, 0, 1, 1])
		if coveredB:
			self.collisions.append([1, 0, 1, 1])
		if coveredC:
			self.collisions.append([0, 1, 1, 1])
		if coveredD:
			self.collisions.append([1, 1, 1, 1])
		
		i = 0
		while i < len(self.collisions):
			c = self.collisions[i]
			c[0] = c[0] * 8 + col * 16
			c[1] = c[1] * 8 + row * 16
			c[2] = c[0] + c[2] * 8
			c[3] = c[1] + c[3] * 8
			self.collisions[i] = c
			i += 1
		
		if self.solid:
			self.isTop = False
			self.isLadder = False
		

class TitleScene:
	def __init__(self):
		self.next = self
		self.flags = ''
		self.flash_counter = 0
		self.options = [
			('start', 'Start Game'),
			('password', "Enter Password"),
			('joystick', "Configure Game Pad")]
		self.index = 0
			
	def processInput(self, events, pressedKeys):
		for event in events:
			if self.flash_counter < 0:
				if event.down:
					if event.action in ['start', 'A', 'B']:
						self.flash_counter = 30
						playNoise('menu_select')
					elif event.action == 'up':
						if self.index > 0:
							self.index -= 1
							playNoise('menu_move')
					elif event.action == 'down':
						if self.index < 2:
							self.index += 1
							playNoise('menu_move')
					
	
	def update(self):
		self.flash_counter -= 1
		if self.flash_counter == 1:
			action = self.options[self.index][0]
			if action == 'start':
				self.next = IntroScene()
			elif action == 'password':
				self.next = PasswordScene()
			elif action == 'joystick':
				self.next = JoystickMenuScene()
			else:
				pass # Assert! or something.
	
	def render(self, screen, renderCounter):
		
		JUKEBOX.ensureSong('title')
		
		screen.fill((0, 0, 0))
		pygame.draw.rect(screen, (0, 128, 255), pygame.Rect(10, 10, 50, 50))
		
		img = renderText((255, 255, 255), "Hello, World!")
		screen.blit(img, (100, 100))
		
		screen.fill((0, 0, 0))
		screen.blit(getImage('slides/title.png'), (0, 0))
		
		showText = True #(renderCounter // 10) % 3 != 0
		if self.flash_counter > 0:
			showText = renderCounter % 4 < 2
		
		index = 0
		y = 136
		left = 64
		
		drawBox(screen, left - 32, y - 16, 24, 9)
		
		for option in self.options:
			
			if index == self.index:
				screen.blit(getText((255, 255, 255), ">"), (left - 16, y))
			
			if index != self.index or showText:
				screen.blit(getText((255, 255, 255), option[1]), (left, y))
			
			index += 1
			y += 16
		
		y += 32
		active_js = getText((128, 128, 128), "Active joystick: ")
		screen.blit(active_js, (8, y))
		x = 8 + active_js.get_width()
		if active_joystick == None:
			js_name = getText((255, 0, 0), "None")
		else:
			js_name = getText((255, 255, 255), active_joystick)
		screen.blit(js_name, (x, y))


WHITESPACE = ' \r\n\t'
def trim(string):
	while len(string) > 0 and string[0] in WHITESPACE:
		string = string[1:]
	while len(string) > 0 and string[-1] in WHITESPACE:
		string = string[:-1]
	return string

def makeGrid(width, height):
	cols = []
	while width > 0:
		cols.append([None] * height)
		width -= 1
	return cols

class StemCell:
	def __init__(self):
		pass
	

def drawBox(screen, x, y, cols, rows):
	left = x
	top = y
	bottom = y + (rows - 1) * 8
	right = x + (cols - 1) * 8
	screen.blit(getImage('misc/border_nw.png'), (left, top))
	screen.blit(getImage('misc/border_ne.png'), (right, top))
	screen.blit(getImage('misc/border_se.png'), (right, bottom))
	screen.blit(getImage('misc/border_sw.png'), (left, bottom))
	x = left + 8
	while x < right:
		screen.blit(getImage('misc/border_n.png'), (x, top))
		screen.blit(getImage('misc/border_s.png'), (x, bottom))
		x += 8
	y = top + 8
	while y < bottom:
		screen.blit(getImage('misc/border_w.png'), (left, y))
		screen.blit(getImage('misc/border_e.png'), (right, y))
		y += 8
	
	pygame.draw.rect(screen, (0, 0, 0), pygame.Rect(left + 8, top + 8, right - left - 8, bottom - top - 8))
	
def legacyMap(fun, things):
	output = []
	for thing in things:
		output.append(fun(thing))
	return output

def abs(x):
	if x < 0: return -x
	return x

class VictorySequenceScene:
	def __init__(self, bg):
		self.next = self
		self.bg = bg
		self.flags = ''
	
		self.images = [
			getImage('slides/close1.png'),
			getImage('slides/close2.png')]
	
		self.player = None
	
		self.counter = 0
	
	def processInput(self, events, pressed):
		pass
	
	def update(self):
		pygame.mixer.music.stop()
		self.counter += 1
	
	def render(self, screen, rc):
		screen.fill(BLACK)
		if self.counter < 60:
			img = self.images[0]
		elif self.counter > 120:
			img = self.images[1]
		else:
			img = self.images[self.counter & 1]
		
		if self.counter == 60:
			playNoise('screen_shaking')
		
		if self.counter == 150:
			self.next = DialogScene(self, 'CloseDialog', False, True, False)
			self.counter += 1
		
		if self.counter == 160:
			self.next = CreditsScene()
		
		screen.blit(img, (0, 16))
	
	

class VolcanoCompleteScene:
	def __init__(self, scene, context):
		self.next = self
		self.flags = ''
		self.bg = scene
		self.context = context
		self.counter = 0
		self.player = scene.player
	
	def processInput(self, events, pressed):
		pass
	
	def update(self):
		self.counter += 1
		if self.counter == 20:
			self.next = DialogScene(self, 'Volcano', False)
		
		if self.counter == 30:
			self.next = PlayScene('ship_1', 8, 9, self.context)
		
	def render(self, screen, rc):
		self.bg.render(screen, rc)
	

FPS = 30
WIDTH = int(256 * 3.4)
HEIGHT = 224 * 3

DEBUG = {'counter':0}
BLACK = (0, 0, 0)
def main():
	pygame.init()

	configureJoysticksFromFile()
	getJoysticksPresent()	
	if joysticks_present.get(previous_active_joystick) != None:
		set_active_joystick(previous_active_joystick)
	real_screen = pygame.display.set_mode((WIDTH, HEIGHT))
	virtual_screen = pygame.Surface((256, 224))
	
	activeScene = OpeningScene()
	renderCounter = 0

	pygame.display.set_caption("Space Squirrel")
	ico = getImage('sprites/space_overworld_right_2.png')
	ico = pygame.transform.scale2x(ico)
	pygame.display.set_icon(ico)
	
	pressedActions = {}
	for action in 'left up down right A B start'.split():
		pressedActions[action] = False
	
	while activeScene != None:
		begin = time.time()
		events = []
		
		poll_active_joystick(events, pressedActions)
		
		quitAttempt = False
		pressedKeys = pygame.key.get_pressed()
		altPressed = pressedKeys[pygame.K_LALT] or pressedKeys[pygame.K_RALT]
		
		for event in pygame.event.get():
			if event.type == pygame.QUIT:
				quitAttempt = True
				
			if event.type == pygame.KEYDOWN and event.key == pygame.K_F4 and altPressed:
				quitAttempt = True
			
			if event.type == pygame.KEYDOWN or event.type == pygame.KEYUP:
				down = event.type == pygame.KEYDOWN
				if event.key == pygame.K_RETURN:
					events.append(MyEvent('start', down))
					pressedActions['start'] = down
				elif event.key == pygame.K_LEFT:
					events.append(MyEvent('left', down))
					pressedActions['left'] = down
				elif event.key == pygame.K_RIGHT:
					events.append(MyEvent('right', down))
					pressedActions['right'] = down
				elif event.key == pygame.K_UP:
					events.append(MyEvent('up', down))
					pressedActions['up'] = down
				elif event.key == pygame.K_DOWN:
					events.append(MyEvent('down', down))
					pressedActions['down'] = down
				elif event.key == pygame.K_SPACE:
					events.append(MyEvent('A', down))
					pressedActions['A'] = down
				elif event.key == pygame.K_b:
					events.append(MyEvent('B', down))
					pressedActions['B'] = down
				elif event.key == pygame.K_F12 and down:
					filename = 1
					while os.path.exists('screencap_' + str(filename) + '.jpg'):
						filename += 1
					pygame.image.save(real_screen, 'screencap_' + str(filename) + '.jpg')
					
		if quitAttempt:
			activeScene = None
			
		if activeScene != None:
			largeScreen = 'W' in activeScene.flags
			
			activeScene.processInput(events, pressedActions)
			activeScene.update()
			if largeScreen:
				activeScene.render(real_screen, renderCounter)
			else:
				activeScene.render(virtual_screen, renderCounter)
			
			activeScene = activeScene.next
			
			if not largeScreen:
				pygame.transform.scale(virtual_screen, (WIDTH, HEIGHT), real_screen)
			
			pygame.display.flip()
		
		renderCounter += 1
		
		end = time.time()
		
		diff = end - begin
		
		delay = 1.0 / FPS - diff
		if delay > 0:
			time.sleep(delay)
		else:
			if SHOW_FRAME_DROP:
				print ("FRAME DROPPING:", 1.0 / diff, "FPS")
		DEBUG['counter'] += 1

main()